import pandas as pd

from cellphonedb.utils import utils


def read_meta_file(path, filename):
    meta_raw = utils.read_data_table_from_file('{}/{}'.format(path, filename), index_column_first=True)
    meta = pd.DataFrame(index=meta_raw.index)
    meta['cell_type'] = meta_raw.iloc[:, 0]
    return meta



from click.testing import CliRunner

from cellphonedb.src.api_endpoints.terminal_api.query_terminal_api_endpoints.query_terminal_commands import \
    get_interaction_gene
from cellphonedb.src.app.flask.flask_app import create_app
from cellphonedb.src.tests.cellphone_flask_test_case import CellphoneFlaskTestCase


class TestTerminalQueryInteractionCalls(CellphoneFlaskTestCase):
    def create_app(self):
        return create_app(raise_non_defined_vars=False, verbose=False)

    def test_call_get_interactions_gene(self):
        runner = CliRunner()
        result = runner.invoke(get_interaction_gene)

        self.assertEqual(0, result.exit_code, 'terminal call get interactions gene error')
        self.assertTrue('ensembl' in str(result.output), 'ensembl column should be present in response')
        self.assertTrue('gene_name' in str(result.output), 'gene_name column shoud be present in response')
        self.assertTrue('hgnc_symbol' in str(result.output), 'hgnc_symbol should be present in response')

    def test_call_get_interactions_gene_withparam(self):
        runner = CliRunner()
        result = runner.invoke(get_interaction_gene, ['--columns=ensembl,gene_name'])

        self.assertEqual(0, result.exit_code, 'terminal call get interactions gene error')
        self.assertTrue('ensembl' in str(result.output), 'ensembl column should be present in response')
        self.assertTrue('gene_name' in str(result.output), 'gene_name column shoud be present in response')
        self.assertTrue('hgnc_symbol' not in str(result.output), 'hgnc_symbol shouldnt be present in response')



from sqlalchemy import MetaData, ForeignKeyConstraint, Table
from sqlalchemy.engine import reflection
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.ddl import DropConstraint, DropTable


# TODO: create sqlalchemy inheritance
class Database:
    def __init__(self, engine):
        self.engine = engine
        self.established_session = None
        self.base = declarative_base()
        self.base_model = None

    @property
    def session(self):
        if not self.established_session:
            session = sessionmaker(bind=self.engine)
            self.established_session = session()

        return self.established_session

    def create_all(self):
        self.base_model.metadata.create_all(self.engine)

    def drop_everything(self):
        conn = self.engine.connect()

        trans = conn.begin()

        inspector = reflection.Inspector.from_engine(self.engine)

        metadata = MetaData()

        tbs = []
        all_fks = []

        for table_name in inspector.get_table_names():
            fks = []
            for fk in inspector.get_foreign_keys(table_name):
                if not fk['name']:
                    continue
                fks.append(
                    ForeignKeyConstraint((), (), name=fk['name'])
                )
            t = Table(table_name, metadata, *fks)
            tbs.append(t)
            all_fks.extend(fks)

        for fkc in all_fks:
            conn.execute(DropConstraint(fkc))

        for table in tbs:
            conn.execute(DropTable(table))

        trans.commit()



import pandas as pd
import numpy as np

def percent_analysis(clusters: dict,
                     threshold: float,
                     interactions: pd.DataFrame,
                     cluster_interactions: list,
                     base_result: pd.DataFrame,
                     separator: str,
                     ) -> pd.DataFrame:
    GENE_ID1 = 'multidata_1_id'
    GENE_ID2 = 'multidata_2_id'

    cluster1_names = cluster_interactions[:, 0]
    cluster2_names = cluster_interactions[:, 1]
    gene1_ids = interactions[GENE_ID1].values
    gene2_ids = interactions[GENE_ID2].values

    x = clusters['percents'].loc[gene1_ids, cluster1_names].values
    y = clusters['percents'].loc[gene2_ids, cluster2_names].values

    result = pd.DataFrame(
        ((x > threshold) * (y > threshold)).astype(int),
        index=interactions.index,
        columns=(pd.Series(cluster1_names) + separator + pd.Series(cluster2_names)).values)

    return result


def get_significant_means(mean_analysis: pd.DataFrame,
                          result_percent: pd.DataFrame) -> pd.DataFrame:
    significant_means = mean_analysis.values.copy()
    mask = result_percent == 0
    significant_means[mask] = np.nan
    return pd.DataFrame(significant_means,
                        index=mean_analysis.index,
                        columns=mean_analysis.columns)


def build_significant_means(mean_analysis: pd.DataFrame,
                            result_percent: pd.DataFrame) -> (pd.Series, pd.DataFrame):
    """
    Calculates the significant means and add rank (number of non empty entries divided by total entries)
    """
    significant_means = get_significant_means(mean_analysis, result_percent)
    significant_mean_rank = significant_means.count(axis=1)  # type: pd.Series
    number_of_clusters = len(significant_means.columns)
    significant_mean_rank = significant_mean_rank.apply(lambda rank: rank / number_of_clusters)
    significant_mean_rank = significant_mean_rank.round(3)
    significant_mean_rank.name = 'rank'
    return significant_mean_rank, significant_means



import os

from cellphonedb.src.app.flask.flask_cellphonedb import CellphonedbFlask

cellphonedb_app = CellphonedbFlask()

file_current_dir = os.path.dirname(os.path.realpath(__file__))
current_execution_dir = os.getcwd()

core_dir = '%s/../core' % file_current_dir
data_dir = '%s/data' % core_dir
output_dir = '%s/out' % current_execution_dir
query_input_dir = '%s' % current_execution_dir
temp_dir = '%s/../../temp' % file_current_dir
output_test_dir = '{}/../tests/out'.format(file_current_dir)
data_test_dir = '{}/../tests/fixtures'.format(file_current_dir)



from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.cellphonedb_app import cellphonedb_app, output_dir


class LocalExporterLauncher(object):
    @staticmethod
    def all():
        LocalExporterLauncher._call_cellphonecore_method(cellphonedb_app.cellphonedb.export.protein)
        LocalExporterLauncher._call_cellphonecore_method(cellphonedb_app.cellphonedb.export.complex)
        LocalExporterLauncher._call_cellphonecore_method(cellphonedb_app.cellphonedb.export.gene)
        LocalExporterLauncher._call_cellphonecore_method(cellphonedb_app.cellphonedb.export.interaction)

    @staticmethod
    def _call_cellphonecore_method(export_method, output_name=None, output_path=None):
        app_logger.debug('Exporting {}'.format(export_method.__name__))
        if not output_name:
            output_name = '%s.csv' % export_method.__name__

        if not output_path:
            output_path = output_dir

        result = export_method()
        result.to_csv('{}/{}'.format(output_path, output_name), index=False)

    def __getattr__(cls, method_name):
        def wrapper(output_name=None, output_path=None):
            export_method = getattr(cellphonedb_app.cellphonedb.export, method_name)
            cls._call_cellphonecore_method(export_method, output_name, output_path)

        return wrapper



import pandas as pd

from cellphonedb.tools.tools_helper import normalize_interactions, interaction_exist


def merge_interactions(interactions_1, interactions_2, interaction_1_key='protein_1', interaction_2_key='protein_2'):
    """
    Merges two interactions dataframens prioritizing keeping first on duplicates
    """

    interactions_2_not_in_1 = interactions_2[
        interactions_2.apply(lambda row: interaction_exist(row, interactions_1, interaction_1_key, interaction_2_key),
                             axis=1) == False]

    interactions = interactions_1.append(interactions_2_not_in_1, sort=True)

    return interactions


def merge_iuphar_imex_interactions(iuphar_interactions: pd.DataFrame,
                                   imex_interactions: pd.DataFrame) -> pd.DataFrame():
    merge_result = merge_interactions(iuphar_interactions, imex_interactions, 'uniprot_1', 'uniprot_2')
    merge_result.fillna({'iuphar': False}, inplace=True)
    return merge_result


# TODO: Make it easier
def merge_iuphar_other_and_curated_interactions(iuphar_other_interactions: pd.DataFrame,
                                                curated_interactions: pd.DataFrame) -> pd.DataFrame:
    all_interactions = iuphar_other_interactions.append(curated_interactions)

    all_interactions.reset_index(inplace=True, drop=True)
    normalized_interactions = normalize_interactions(all_interactions, 'partner_a', 'partner_b')

    duplicated_interactions = normalized_interactions[
        normalized_interactions.duplicated(['partner_a', 'partner_b'], keep=False)]

    unique_interactions = duplicated_interactions.drop_duplicates(['partner_a', 'partner_b'])

    def merge_values(interaction: pd.Series) -> pd.Series:
        interaction = interaction.copy()
        duplicated = duplicated_interactions[
            (duplicated_interactions['partner_a'] == interaction['partner_a']) & (
                    duplicated_interactions['partner_b'] == interaction['partner_b'])]

        if not duplicated[duplicated['annotation_strategy'] == 'curated'].empty:
            interaction = duplicated[duplicated['annotation_strategy'] == 'curated'].iloc[0]

        if duplicated['iuphar'].any():
            interaction['iuphar'] = True

        return interaction

    merged_duplicated_interactions = unique_interactions.apply(merge_values, axis=1)

    non_repeated_interactions = normalized_interactions.drop_duplicates(['partner_a', 'partner_b'],
                                                                        keep=False)
    interactions_merged = non_repeated_interactions.append(merged_duplicated_interactions, sort=True, ignore_index=True)
    interactions_merged.fillna({'iuphar': False}, inplace=True)
    interactions_merged = interactions_merged.astype(
        {'iuphar': bool})

    return interactions_merged



import warnings
warnings.simplefilter("ignore", UserWarning)

import itertools
from functools import partial
from multiprocessing.pool import Pool

import pandas as pd
import numpy as np
import numpy_groupies as npg
from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.models.complex import complex_helper


def get_significant_means(real_mean_analysis: pd.DataFrame,
                          result_percent: pd.DataFrame,
                          min_significant_mean: float) -> pd.DataFrame:
    """
    If the result_percent value is > min_significant_mean, sets the value to NaN, else, sets the mean value.

    EXAMPLE:
    INPUT:

    real mean
                cluster1    cluster2    cluster
    ensembl1    0.1         1.0         2.0
    ensembl2    2.0         0.1         0.2
    ensembl3    0.3         0.0         0.5

    result percent

                cluster1    cluster2    cluster
    ensembl1    0.0         1.0         1.0
    ensembl2    0.04        0.03        0.62
    ensembl3    0.3         0.55        0.02

    min_significant_mean = 0.05

    RESULT:

                cluster1    cluster2    cluster
    ensembl1    0.1         NaN         NaN
    ensembl2    2.0         0.1         NaN
    ensembl3    NaN         NaN         0.5
    """
    significant_means = real_mean_analysis.values.copy()
    mask = result_percent > min_significant_mean
    significant_means[mask] = np.nan
    return pd.DataFrame(significant_means,
                        index=real_mean_analysis.index,
                        columns=real_mean_analysis.columns)


def shuffle_meta(meta: pd.DataFrame) -> pd.DataFrame:
    """
    Permutates the meta values aleatory generating a new meta file
    """
    meta_copy = meta.copy()
    np.random.shuffle(meta_copy['cell_type'].values)

    return meta_copy


def build_clusters(meta: pd.DataFrame,
                   counts: pd.DataFrame,
                   complex_composition: pd.DataFrame,
                   skip_percent: bool) -> dict:
    """
    Builds a cluster structure and calculates the means values
    """
    CELL_TYPE = 'cell_type'
    COMPLEX_ID = 'complex_multidata_id'
    PROTEIN_ID = 'protein_multidata_id'

    meta[CELL_TYPE] = meta[CELL_TYPE].astype('category')
    cluster_names = meta[CELL_TYPE].cat.categories

    # Simple genes cluster counts
    cluster_means = pd.DataFrame(
        npg.aggregate(meta[CELL_TYPE].cat.codes, counts.values, func='mean', axis=1),
        index=counts.index,
        columns=cluster_names.to_list()
    )
    if not skip_percent:
        cluster_pcts = pd.DataFrame(
            npg.aggregate(meta[CELL_TYPE].cat.codes, (counts > 0).astype(int).values, func='mean', axis=1),
            index=counts.index,
            columns=cluster_names.to_list()
        )
    else:
        cluster_pcts = pd.DataFrame(index=counts.index, columns=cluster_names.to_list())

    # Complex genes cluster counts
    if not complex_composition.empty:
        complexes = complex_composition.groupby(COMPLEX_ID).apply(lambda x: x[PROTEIN_ID].values).to_dict()
        complex_cluster_means = pd.DataFrame(
            {complex_id: cluster_means.loc[protein_ids].min(axis=0).values
             for complex_id, protein_ids in complexes.items()},
            index=cluster_means.columns
        ).T
        cluster_means = cluster_means.append(complex_cluster_means)
        if not skip_percent:
            complex_cluster_pcts = pd.DataFrame(
                {complex_id: cluster_pcts.loc[protein_ids].min(axis=0).values
             for complex_id, protein_ids in complexes.items()},
                index=cluster_pcts.columns
            ).T
            cluster_pcts = cluster_pcts.append(complex_cluster_pcts)

    return {'names': cluster_names, 'means': cluster_means, 'percents': cluster_pcts}


def filter_counts_by_interactions(counts: pd.DataFrame,
                                  interactions: pd.DataFrame) -> pd.DataFrame:
    """
    Removes counts if is not defined in interactions components
    """
    multidata_genes_ids = interactions['multidata_1_id'].append(
        interactions['multidata_2_id']).drop_duplicates().tolist()

    counts_filtered = counts.filter(multidata_genes_ids, axis=0)

    return counts_filtered


def filter_empty_cluster_counts(counts: pd.DataFrame) -> pd.DataFrame:
    """
    Remove count with all values to zero
    """
    if counts.empty:
        return counts

    filtered_counts = counts[counts.apply(lambda row: row.sum() > 0, axis=1)]
    return filtered_counts


def mean_pvalue_result_build(real_mean_analysis: pd.DataFrame, result_percent: pd.DataFrame,
                             interactions_data_result: pd.DataFrame) -> pd.DataFrame:
    """
    Merges the pvalues and means in one table
    """
    mean_pvalue_result = pd.DataFrame(index=real_mean_analysis.index)
    for interaction_cluster in real_mean_analysis.columns.values:
        mean_pvalue_result[interaction_cluster] = real_mean_analysis[interaction_cluster].astype(str).str.cat(
            result_percent[interaction_cluster].astype(str), sep=' | ')

    mean_pvalue_result = pd.concat([interactions_data_result, mean_pvalue_result], axis=1, join='inner', sort=False)

    return mean_pvalue_result


def get_cluster_combinations(cluster_names: np.array) -> np.array:
    """
    Calculates and sort combinations including itself

    ie

    INPUT
    cluster_names = ['cluster1', 'cluster2', 'cluster3']

    RESULT
    [('cluster1','cluster1'),('cluster1','cluster2'),('cluster1','cluster3'),
     ('cluster2','cluster1'),('cluster2','cluster2'),('cluster2','cluster3'),
     ('cluster3','cluster1'),('cluster3','cluster2'),('cluster3','cluster3')]

    """
    return np.array(np.meshgrid(cluster_names.values, cluster_names.values)).T.reshape(-1, 2)


def build_result_matrix(interactions: pd.DataFrame, cluster_interactions: list, separator: str) -> pd.DataFrame:
    """
    builds an empty cluster matrix to fill it later
    """
    columns = []

    for cluster_interaction in cluster_interactions:
        columns.append('{}{}{}'.format(cluster_interaction[0], separator, cluster_interaction[1]))

    result = pd.DataFrame(index=interactions.index, columns=columns, dtype=float)

    return result


def mean_analysis(interactions: pd.DataFrame,
                  clusters: dict,
                  cluster_interactions: list,
                  base_result: pd.DataFrame,
                  separator: str) -> pd.DataFrame:
    """
    Calculates the mean for the list of interactions and for each cluster

    sets 0 if one of both is 0

    EXAMPLE:
        cluster_means
                   cluster1    cluster2    cluster3
        ensembl1     0.0         0.2         0.3
        ensembl2     0.4         0.5         0.6
        ensembl3     0.7         0.0         0.9

        interactions:

        ensembl1,ensembl2
        ensembl2,ensembl3

        RESULT:
                              cluster1_cluster1   cluster1_cluster2   ...   cluster3_cluster2   cluster3_cluster3
        ensembl1_ensembl2     mean(0.0,0.4)*      mean(0.0,0.5)*            mean(0.3,0.5)       mean(0.3,0.6)
        ensembl2_ensembl3     mean(0.4,0.7)       mean(0.4,0.0)*            mean(0.6,0.0)*      mean(0.6,0.9)


        results with * are 0 because one of both components is 0.
    """
    GENE_ID1 = 'multidata_1_id'
    GENE_ID2 = 'multidata_2_id'

    cluster1_names = cluster_interactions[:, 0]
    cluster2_names = cluster_interactions[:, 1]
    gene1_ids = interactions[GENE_ID1].values
    gene2_ids = interactions[GENE_ID2].values

    x = clusters['means'].loc[gene1_ids, cluster1_names].values
    y = clusters['means'].loc[gene2_ids, cluster2_names].values

    result = pd.DataFrame(
        (x > 0) * (y > 0) * (x + y) / 2,
        index=interactions.index,
        columns=(pd.Series(cluster1_names) + separator + pd.Series(cluster2_names)).values)

    return result


def percent_analysis(clusters: dict,
                     threshold: float,
                     interactions: pd.DataFrame,
                     cluster_interactions: list,
                     base_result: pd.DataFrame,
                     separator: str) -> pd.DataFrame:
    """
    Calculates the percents for cluster interactions and foreach gene interaction

    If one of both is not 0 sets the value to 0. Else sets 1

    EXAMPLE:
        INPUT:

        threshold = 0.1
        cluster1 = cell1,cell2
        cluster2 = cell3

                     cell1       cell2      cell3
        ensembl1     0.0         0.6         0.3
        ensembl2     0.1         0.05         0.06
        ensembl3     0.0         0.0         0.9

        interactions:

        ensembl1,ensembl2
        ensembl1,ensembl3


        (after percents calculation)

                     cluster1    cluster2
        ensembl1     0           0
        ensembl2     1           1
        ensembl3     1           0

        RESULT:
                            cluster1_cluster1   cluster1_cluster2   cluster2_cluster1   cluster2_cluster2
        ensembl1_ensembl2   (0,1)-> 0           (0,1)-> 0           (0,1)->0            (0,1)->0
        ensembl1_ensembl3   (0,1)-> 0           (0,0)-> 1           (0,1)->0            (0,0)->1


    """
    GENE_ID1 = 'multidata_1_id'
    GENE_ID2 = 'multidata_2_id'

    cluster1_names = cluster_interactions[:, 0]
    cluster2_names = cluster_interactions[:, 1]
    gene1_ids = interactions[GENE_ID1].values
    gene2_ids = interactions[GENE_ID2].values

    x = clusters['percents'].loc[gene1_ids, cluster1_names].values
    y = clusters['percents'].loc[gene2_ids, cluster2_names].values

    result = pd.DataFrame(
        ((x > threshold) * (y > threshold)).astype(int),
        index=interactions.index,
        columns=(pd.Series(cluster1_names) + separator + pd.Series(cluster2_names)).values)

    return result


def shuffled_analysis(iterations: int,
                      meta: pd.DataFrame,
                      counts: pd.DataFrame,
                      interactions: pd.DataFrame,
                      cluster_interactions: list,
                      complex_composition: pd.DataFrame,
                      real_mean_analysis: pd.DataFrame,
                      base_result: pd.DataFrame,
                      threads: int,
                      separator: str) -> list:
    """
    Shuffles meta and calculates the means for each and saves it in a list.

    Runs it in a multiple threads to run it faster
    """
    with Pool(processes=threads) as pool:
        statistical_analysis_thread = partial(_statistical_analysis,
                                              base_result,
                                              cluster_interactions,
                                              counts,
                                              interactions,
                                              meta,
                                              complex_composition,
                                              separator,
                                              real_mean_analysis)
        results = pool.map(statistical_analysis_thread, range(iterations))

    return results


def _statistical_analysis(base_result,
                          cluster_interactions,
                          counts,
                          interactions,
                          meta,
                          complex_composition: pd.DataFrame,
                          separator,
                          real_mean_analysis: pd.DataFrame,
                          iteration_number) -> pd.DataFrame:
    """
    Shuffles meta dataset and calculates calculates the means
    """
    shuffled_meta = shuffle_meta(meta)
    shuffled_clusters = build_clusters(shuffled_meta,
                                       counts,
                                       complex_composition,
                                       skip_percent=True)

    shuffled_mean_analysis = mean_analysis(interactions,
                                           shuffled_clusters,
                                           cluster_interactions,
                                           base_result,
                                           separator)

    result_mean_analysis = shuffled_greater_than_real(shuffled_mean_analysis,
                                                    real_mean_analysis)
    return result_mean_analysis


def shuffled_greater_than_real(shuffled_mean_analysis: pd.DataFrame,
                             real_mean_analysis: pd.DataFrame):
    return np.packbits(shuffled_mean_analysis.values > real_mean_analysis.values, axis=None)


def build_percent_result(real_mean_analysis: pd.DataFrame, real_percents_analysis: pd.DataFrame,
                         statistical_mean_analysis: list, interactions: pd.DataFrame, cluster_interactions: list,
                         base_result: pd.DataFrame, separator: str) -> pd.DataFrame:
    """
    Calculates the pvalues after statistical analysis.

    If real_percent or real_mean are zero, result_percent is 1

    If not:
    Calculates how many shuffled means are bigger than real mean and divides it for the number of
    the total iterations

    EXAMPLE:
        INPUT:

        real_mean_analysis:
                      cluster1_cluster1   cluster1_cluster2 ...
        interaction1  0.5                 0.4
        interaction2  0.0                 0.2


        real_percents_analysis:
                      cluster1_cluster1   cluster1_cluster2 ...
        interaction1  1                   0
        interaction2  0                   1

        statistical means:
        [
                        cluster1_cluster1   cluster1_cluster2 ...
        interaction1    0.6                 0.1
        interaction2    0.0                 0.2

        ,
                      cluster1_cluster1   cluster1_cluster2 ...
        interaction1  0.5                 0.4
        interaction2  0.0                 0.6
        ]

        iterations = 2


        RESULT:

                        cluster1_cluster1   cluster1_cluster2 ...
        interaction1    1                   1
        interaction2    1                   0.5


    """
    core_logger.info('Building Pvalues result')
    percent_result = np.zeros(real_mean_analysis.shape)
    result_size = percent_result.size
    result_shape = percent_result.shape

    for statistical_mean in statistical_mean_analysis:
        percent_result += np.unpackbits(statistical_mean, axis=None)[:result_size].reshape(result_shape)
    percent_result /= len(statistical_mean_analysis)

    mask = (real_mean_analysis.values == 0) | (real_percents_analysis == 0)

    percent_result[mask] = 1

    return pd.DataFrame(percent_result, index=base_result.index, columns=base_result.columns)


def interacting_pair_build(interactions: pd.DataFrame) -> pd.Series:
    """
    Returns the interaction result formated with prefixes
    """

    def get_interactor_name(interaction: pd.Series, suffix: str) -> str:
        if interaction['is_complex{}'.format(suffix)]:
            return interaction['name{}'.format(suffix)]

        return interaction['gene_name{}'.format(suffix)]

    interacting_pair = interactions.apply(
        lambda interaction: '{}_{}'.format(get_interactor_name(interaction, '_1'),
                                           get_interactor_name(interaction, '_2')), axis=1)

    interacting_pair.rename('interacting_pair', inplace=True)

    return interacting_pair


def build_significant_means(real_mean_analysis: pd.DataFrame,
                            result_percent: pd.DataFrame,
                            min_significant_mean: float) -> (pd.Series, pd.DataFrame):
    """
    Calculates the significant means and add rank (number of non empty entries divided by total entries)
    """
    significant_means = get_significant_means(real_mean_analysis, result_percent, min_significant_mean)
    significant_mean_rank = significant_means.count(axis=1)  # type: pd.Series
    number_of_clusters = len(significant_means.columns)
    significant_mean_rank = significant_mean_rank.apply(lambda rank: rank / number_of_clusters)
    significant_mean_rank = significant_mean_rank.round(3)
    significant_mean_rank.name = 'rank'
    return significant_mean_rank, significant_means


def filter_interactions_by_counts(interactions: pd.DataFrame,
                                  counts: pd.DataFrame,
                                  complex_composition: pd.DataFrame) -> pd.DataFrame:
    multidatas = list(counts.index)

    if not complex_composition.empty:
        multidatas += complex_composition['complex_multidata_id'].to_list() + complex_composition[
            'protein_multidata_id'].to_list()

    multidatas = list(set(multidatas))

    def filter_interactions(interaction: pd.Series) -> bool:
        if interaction['multidata_1_id'] in multidatas and interaction['multidata_2_id'] in multidatas:
            return True
        return False

    interactions_filtered = interactions[interactions.apply(filter_interactions, axis=1)]

    return interactions_filtered


def prefilters(interactions: pd.DataFrame,
               counts: pd.DataFrame,
               complexes: pd.DataFrame,
               complex_compositions: pd.DataFrame):
    """
    - Finds the complex defined in counts and calculates their counts values
    - Remove interactions if the simple component ensembl is not in the counts list
    - Remove interactions if the complex component is not in the calculated complex list
    - Remove undefined simple counts
    - Merge simple filtered counts and calculated complex counts
    - Remove duplicated counts
    """
    counts_filtered = filter_empty_cluster_counts(counts)
    complex_composition_filtered, counts_complex = get_involved_complex_from_counts(counts_filtered,
                                                                                    complex_compositions)

    interactions_filtered = filter_interactions_by_counts(interactions,
                                                          counts_filtered,
                                                          complex_composition_filtered)

    counts_simple = filter_counts_by_interactions(counts_filtered, interactions_filtered)

    counts_filtered = counts_simple.append(counts_complex, sort=False)
    counts_filtered = counts_filtered[~counts_filtered.index.duplicated()]

    return interactions_filtered, counts_filtered, complex_composition_filtered

def get_involved_complex_from_counts(counts: pd.DataFrame,
                                     complex_composition: pd.DataFrame) -> (
        pd.DataFrame, pd.DataFrame):
    """
    Finds the complexes defined in counts and calculates the counts values
    """
    proteins_in_complexes = complex_composition['protein_multidata_id'].drop_duplicates().tolist()

    # Remove counts that can't be part of a complex
    counts_filtered = counts[
        counts.apply(lambda count: count.name in proteins_in_complexes, axis=1)]

    # Find complexes with all components defined in counts
    complex_composition_filtered = complex_helper.get_involved_complex_composition_from_protein(counts_filtered,
                                                                                                complex_composition)

    if complex_composition_filtered.empty:
        return complex_composition_filtered, pd.DataFrame(columns=counts.columns)

    available_complex_proteins = complex_composition_filtered['protein_multidata_id'].drop_duplicates().to_list()

    # Remove counts that are not defined in selected complexes
    counts_filtered = counts_filtered[
        counts_filtered.apply(lambda count: count.name in available_complex_proteins, axis=1)]

    return complex_composition_filtered, counts_filtered



import pandas as pd

from cellphonedb.src.core.models.interaction import interaction_properties
from cellphonedb.utils import unique_id_generator


def call(interactions: pd.DataFrame, multidatas: pd.DataFrame) -> pd.DataFrame:
    interactions_processed = _set_interactor_property(interactions, multidatas)

    return interactions_processed


def _set_interactor_property(interactions_processed: pd.DataFrame,
                             multidatas: pd.DataFrame) -> pd.DataFrame:
    interactions_processed_expanded = pd.merge(interactions_processed, multidatas, left_on=['partner_a'],
                                               right_on=['name'])
    interactions_processed_expanded = pd.merge(interactions_processed_expanded, multidatas,
                                               left_on=['partner_b'], right_on=['name'],
                                               suffixes=['_x', '_y'])
    interactions_processed_expanded = interactions_processed_expanded[
        interactions_processed_expanded.apply(
            lambda interaction: interaction_properties.is_cellphonedb_interactor(interaction, ('_x', '_y')), axis=1)]

    interactions_processed_expanded['id_cp_interaction'] = interactions_processed_expanded.apply(
        lambda interaction: unique_id_generator.interaction(interaction, ('_x', '_y')), axis=1)

    return interactions_processed_expanded



import pandas as pd
import numpy as np
from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.core.generators.generator_helper import set_defaults
from cellphonedb.src.exceptions.MissingRequiredColumns import MissingRequiredColumns


def complex_generator(curated_complex: pd.DataFrame, user_complex: pd.DataFrame, log_path: str) -> pd.DataFrame:
    if (isinstance(user_complex, pd.DataFrame) and user_complex.empty) or \
            (not isinstance(user_complex, pd.DataFrame) and not user_complex):
        return curated_complex

    try:
        return _merge_complex(curated_complex, user_complex, log_path)
    except MissingRequiredColumns as e:
        app_logger.error(e)


def _merge_complex(base_complex: pd.DataFrame, additional: pd.DataFrame, log_file: str) -> pd.DataFrame:
    additional = additional.copy()

    defaults = {
        'uniprot_3': np.nan,
        'uniprot_4': np.nan,
        'receptor': False,
        'integrin': False,
        'other': False,
        'other_desc': np.nan,
        'peripheral': False,
        'receptor_desc': np.nan,
        'secreted_desc': np.nan,
        'secreted_highlight': False,
        'secreted': False,
        'transmembrane': False,
        'pdb_structure': False,
        'pdb_id': np.nan,
        'stoichiometry': np.nan,
        'comments_complex': np.nan
    }

    default_types = {
        'complex_name': str,
        'uniprot_1': str,
        'uniprot_2': str,
        'uniprot_3': str,
        'uniprot_4': str,
        'transmembrane': bool,
        'peripheral': bool,
        'secreted': bool,
        'secreted_desc': str,
        'secreted_highlight': bool,
        'receptor': bool,
        'receptor_desc': str,
        'integrin': bool,
        'other': bool,
        'other_desc': str,
        'pdb_id': str,
        'pdb_structure': str,
        'stoichiometry': str,
        'comments_complex': str,
    }

    result_columns = list(default_types.keys())

    required_columns = ['complex_name', 'uniprot_1', 'uniprot_2']

    if not set(required_columns).issubset(additional):
        raise MissingRequiredColumns(list(set(required_columns).difference(additional)))

    # TODO: Fill NA
    # Here we set defaults for additional data
    additional = set_defaults(additional, defaults)

    # we will only use these columns
    additional = additional[result_columns]
    base_complex = base_complex[result_columns]

    # Type casting to ensure they are equal
    base_complex = base_complex.astype(default_types)
    additional = additional.astype(default_types)

    join_key = 'complex_name'

    merged_complex = base_complex.append(additional, ignore_index=True, sort=False).drop_duplicates()

    if merged_complex.duplicated(join_key).any():
        app_logger.warning('There are differences between merged files: logged to {}'.format(log_file))

        log = merged_complex[merged_complex.duplicated(join_key, keep=False)].sort_values(join_key)
        log.to_csv(log_file, index=False, sep='\t')

    merged_complex.drop_duplicates(join_key, keep='last', inplace=True)
    return merged_complex



from typing import Callable

import click

from cellphonedb.src.api_endpoints.terminal_api.util.choose_database import choose_database
from cellphonedb.src.app import cpdb_app
from cellphonedb.src.local_launchers.local_query_launcher import LocalQueryLauncher


def common_options(f: Callable) -> Callable:
    options = [
        click.option('--verbose/--quiet', default=True, help='Print or hide cellphonedb logs [verbose]'),
        click.option('--database', default='latest', callback=choose_database),
    ]

    for option in reversed(options):
        f = option(f)

    return f


@click.command()
@click.argument('element')
@common_options
def find_interactions_by_element(element: str, verbose: bool, database: str):
    LocalQueryLauncher(cpdb_app.create_app(verbose, database)).find_interactions_by_element(element)


@click.command()
@click.option('--columns', default=None, help='Columns to set in the result')
@common_options
def get_interaction_gene(columns: str, verbose: bool, database: str):
    LocalQueryLauncher(cpdb_app.create_app(verbose, database)).get_interaction_gene(columns)


@click.command()
@click.argument('partial_element')
@common_options
def autocomplete(partial_element: str, verbose: bool, database: str) -> None:
    LocalQueryLauncher(cpdb_app.create_app(verbose, database)).autocomplete_element(partial_element)



import pandas as pd

from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.database.Repository import Repository
from cellphonedb.src.core.database.sqlalchemy_models.db_model_complex import Complex
from cellphonedb.src.core.database.sqlalchemy_models.db_model_complex_composition import ComplexComposition
from cellphonedb.src.core.database.sqlalchemy_models.db_model_gene import Gene
from cellphonedb.src.core.database.sqlalchemy_models.db_model_multidata import Multidata
from cellphonedb.src.core.database.sqlalchemy_models.db_model_protein import Protein
from cellphonedb.src.core.utils import filters


class ComplexRepository(Repository):
    name = 'complex'

    def get_all(self) -> pd.DataFrame:
        query = self.database_manager.database.session.query(Complex)
        result = pd.read_sql(query.statement, self.database_manager.database.engine)

        return result

    def get_all_expanded(self) -> pd.DataFrame:
        query = self.database_manager.database.session.query(Complex, Multidata).join(Multidata)
        result = pd.read_sql(query.statement, self.database_manager.database.engine)

        return result

    def get_all_compositions(self) -> pd.DataFrame:
        query = self.database_manager.database.session.query(ComplexComposition)
        result = pd.read_sql(query.statement, self.database_manager.database.engine)

        return result

    def get_all_compositions_expanded(self, include_gene: bool = True) -> pd.DataFrame:
        query = self.database_manager.database.session.query(ComplexComposition)
        complex_composition = pd.read_sql(query.statement, self.database_manager.database.engine)

        protein_gene_join = Protein.protein_multidata_id == Multidata.id_multidata
        if include_gene:
            gene_protein_join = Gene.protein_id == Protein.id_protein
            multidatas_proteins_query = self.database_manager.database.session.query(Gene,
                                                                                     Protein,
                                                                                     Multidata).join(Protein,
                                                                                                     gene_protein_join
                                                                                                     ).join(Multidata,
                                                                                                            protein_gene_join)
        else:
            multidatas_proteins_query = self.database_manager.database.session.query(Protein, Multidata).join(
                Multidata, protein_gene_join)

        multidatas_proteins = pd.read_sql(multidatas_proteins_query.statement, self.database_manager.database.engine)
        multidatas_proteins.columns = multidatas_proteins.columns.map(lambda column: column + '_protein')

        multidatas_complexes_query = self.database_manager.database.session.query(Multidata)
        multidatas_complexes = pd.read_sql(multidatas_complexes_query.statement, self.database_manager.database.engine)
        multidatas_complexes.columns = multidatas_complexes.columns.map(lambda column: column + '_complex')

        complex_composition_expanded = pd.merge(complex_composition, multidatas_complexes,
                                                left_on='complex_multidata_id',
                                                right_on='id_multidata_complex')

        complex_composition_expanded = pd.merge(complex_composition_expanded, multidatas_proteins,
                                                left_on='protein_multidata_id',
                                                right_on='id_multidata_protein')

        return complex_composition_expanded

    def get_complex_by_multidatas(self, multidatas: pd.DataFrame, all_proteins_expressed: bool = True) -> pd.DataFrame:
        complex_composition = self.get_all_compositions()

        multidatas_ids = multidatas['id_multidata'].to_frame()
        complex_composition_merged = pd.merge(complex_composition, multidatas_ids, left_on='protein_multidata_id',
                                              right_on='id_multidata')

        if complex_composition_merged.empty:
            return complex_composition_merged

        def all_protein_expressed(complex):
            number_proteins_in_counts = len(
                complex_composition_merged[
                    complex_composition_merged['complex_multidata_id'] == complex['complex_multidata_id']])

            if number_proteins_in_counts < complex['total_protein']:
                return False

            return True

        if all_proteins_expressed:
            complex_composition_merged = complex_composition_merged[
                complex_composition_merged.apply(all_protein_expressed, axis=1)]

        complexes = self.get_all_expanded()
        complex_composition_merged = pd.merge(complex_composition_merged, complexes,
                                              left_on='complex_multidata_id',
                                              right_on='id_multidata',
                                              suffixes=['_protein', ''])

        complex_composition_merged.drop_duplicates(['complex_multidata_id'], inplace=True)

        return complex_composition_merged

    # TODO: it needs to be refactored
    def add(self, complexes):
        """
        Uploads complex data from csv.

        - Creates new complexes in Multidata table
        - Creates reference in Complex table
        - Creates complex composition to define complexes.
        """

        if complexes.empty:
            return

        existing_complexes = self.database_manager.database.session.query(Multidata.name).all()
        existing_complexes = [c[0] for c in existing_complexes]
        proteins = self.database_manager.database.session.query(Multidata.name, Multidata.id_multidata).join(
            Protein).all()
        proteins = {p[0]: p[1] for p in proteins}

        # Get complex composition info
        complete_indices = []
        incomplete_indices = []
        missing_proteins = []
        complex_map = {}
        for index, row in complexes.iterrows():
            missing = False
            protein_id_list = []
            for protein in ['protein_1', 'protein_2',
                            'protein_3', 'protein_4']:
                if not pd.isnull(row[protein]):
                    protein_id = proteins.get(row[protein])
                    if protein_id is None:
                        missing = True
                        missing_proteins.append(row[protein])
                    else:
                        protein_id_list.append(protein_id)
            if not missing:
                complex_map[row['name']] = protein_id_list
                complete_indices.append(int(index))
            else:
                incomplete_indices.append(int(index))

        if len(incomplete_indices) > 0:
            core_logger.warning('MISSING PROTEINS:')
            for protein in missing_proteins:
                core_logger.warning('MISSING PROTEINS:')
                core_logger.warning(protein)

            core_logger.warning('COMEPLEXES WITH MISSING PROTEINS:')
            incomplete_complexes = complexes.iloc[incomplete_indices, :]['name']
            for incomplete_complex in incomplete_complexes:
                core_logger.warning(incomplete_complex)

        # Insert complexes
        if not complexes.empty:
            # Remove unwanted columns
            removal_columns = list(
                [x for x in complexes.columns if 'protein_' in x or 'Name_' in x or 'Unnamed' in x])
            # removal_columns += ['comments']
            complexes.drop(removal_columns, axis=1, inplace=True)

            # Remove rows with missing complexes
            complexes = complexes.iloc[complete_indices, :]

            # Convert ints to bool
            bools = ['receptor', 'other', 'secreted_highlight', 'transmembrane', 'secreted',
                     'peripheral']
            complexes[bools] = complexes[bools].astype(bool)

            # Drop existing complexes
            complexes = complexes[complexes['name'].apply(
                lambda x: x not in existing_complexes)]

            multidata_df = filters.remove_not_defined_columns(complexes.copy(),
                                                              self.database_manager.get_column_table_names(
                                                                  'multidata_table'))

            multidata_df = self._add_complex_optimitzations(multidata_df)
            multidata_df.to_sql(name='multidata_table', if_exists='append', con=self.database_manager.database.engine,
                                index=False, chunksize=50)

        # Now find id's of new complex rows
        new_complexes = self.database_manager.database.session.query(Multidata.name, Multidata.id_multidata).all()
        new_complexes = {c[0]: c[1] for c in new_complexes}

        # Build set of complexes
        complex_set = []
        complex_table = []
        for complex_name in complex_map:
            complex_id = new_complexes[complex_name]
            for protein_id in complex_map[complex_name]:
                complex_set.append((complex_id, protein_id, len(complex_map[complex_name])))
            complex_table.append({'complex_multidata_id': complex_id, 'name': complex_name})

        # Insert complex composition
        complex_set_df = pd.DataFrame(complex_set,
                                      columns=['complex_multidata_id', 'protein_multidata_id', 'total_protein'])

        complex_table_df = pd.DataFrame(complex_table)
        complex_table_df = pd.merge(complex_table_df, complexes, on='name')

        filters.remove_not_defined_columns(complex_table_df,
                                           self.database_manager.get_column_table_names('complex_table'))

        complex_table_df.to_sql(
            name='complex_table', if_exists='append',
            con=self.database_manager.database.engine, index=False, chunksize=50)

        complex_set_df.to_sql(
            name='complex_composition_table', if_exists='append',
            con=self.database_manager.database.engine, index=False, chunksize=50)

    @staticmethod
    def _add_complex_optimitzations(multidatas):
        multidatas['is_complex'] = True

        return multidatas



from flask_restful import Api

from cellphonedb.src.api_endpoints.web_api.query_web_api_endpoints import query_web_api_routes


def add(api: Api, prefix=''):
    query_web_api_routes.add(api, prefix=prefix + '/query')



import pandas as pd


def remove_not_defined_columns(data_frame: pd.DataFrame, defined_columns: list) -> pd.DataFrame:
    """

    :type data_frame: pd.DataFrame
    :type defined_columns: list
    :rtype: pd.DataFrame
    """
    data_frame_keys = list(data_frame.keys())

    for key in data_frame_keys:
        if key not in defined_columns:
            data_frame.drop(key, axis=1, inplace=True)

    return data_frame



import pandas as pd

from cellphonedb.utils import dataframe_format


def dataframes_has_same_data(dataframe1: pd.DataFrame, dataframe2: pd.DataFrame,
                             round_decimals: bool = False) -> pd.DataFrame:
    dataframe1 = dataframe1.copy(deep=True)
    dataframe2 = dataframe2.copy(deep=True)

    columns_names_1 = list(dataframe1.columns.values)
    columns_names_1.sort()

    dataframe1 = dataframe_format.bring_columns_to_end(columns_names_1, dataframe1)

    columns_names_2 = list(dataframe2.columns.values)
    columns_names_2.sort()
    dataframe2 = dataframe_format.bring_columns_to_end(columns_names_2, dataframe2)

    if not dataframe1.empty:
        dataframe1 = dataframe1.sort_values(columns_names_1).reset_index(drop=True)

        if round_decimals:
            dataframe1 = dataframe1.round(5)

    if not dataframe2.empty:
        dataframe2 = dataframe2.sort_values(columns_names_2).reset_index(drop=True)

        if round_decimals:
            dataframe2 = dataframe2.round(5)

    if dataframe1.empty and dataframe2.empty:
        return pd.Series(dataframe1.columns.values).equals(pd.Series(dataframe2.columns.values))

    return dataframe1.equals(dataframe2)



from unittest import TestCase

import os

import pandas as pd

from cellphonedb.utils import utils


class TestOpenDataFile(TestCase):

    def test_extensions(self):
        extensions = ['csv', 'tsv', 'txt']
        base_name = 'example_data'
        for extension in extensions:
            self.assert_open_file(base_name, extension, False, '')

    def test_custom(self):
        self.assert_open_file('example_data_custom', 'csv', False, ';')

    def assert_open_file(self, base_name, extension, index_column_first, separator):
        fixtures_dir = '{}/fixtures'.format(self.current_dir)
        result = utils.read_data_table_from_file('{}/{}.{}'.format(fixtures_dir, base_name, extension),
                                                 index_column_first, separator)
        expected_result = pd.read_csv('{}/example_data.csv'.format(fixtures_dir))

        self.assertTrue(result.equals(expected_result))

    @property
    def current_dir(self):
        return os.path.dirname(os.path.realpath(__file__))



import pandas as pd


def gene_generator(ensembl_db: pd.DataFrame,
                   uniprot_db: pd.DataFrame,
                   hla_genes: pd.DataFrame,
                   user_gene: pd.DataFrame,
                   result_columns: list) -> pd.DataFrame:
    def get_first_gene_name(gene_names: str) -> str:
        if type(gene_names) != str:
            return ''
        gene_names = gene_names.split(' ')

        return gene_names[0]

    uniprot_db = uniprot_db.copy()
    uniprot_db.dropna(inplace=True)
    uniprot_db['gene_name'] = uniprot_db['gene_names'].apply(get_first_gene_name)
    uniprot_db.drop_duplicates(['gene_name', 'uniprot'], inplace=True)

    # Remove hla genes
    no_hla_uniprots = uniprot_db[~uniprot_db['gene_name'].str.contains('HLA')]

    # Merge with ensembl database
    cpdb_genes = no_hla_uniprots.merge(ensembl_db, how='inner', on='gene_name',
                                                     sort=False, suffixes=('', '_ensembl')).drop_duplicates(
        ['ensembl', 'uniprot', 'gene_name'])

    # Add additional non-repeted  ensembl-gene_names based on uniprot
    ensembl_db_filtered = ensembl_db.drop_duplicates()
    ensembl_db_filtered.dropna(inplace=True)

    # print('duplicated ensembl in ensembl list')
    # print(len(ensembl_db_filtered[ensembl_db_filtered['ensembl'].duplicated()]))

    # Add only if the uniprot exist in result gene list
    additional_genes = ensembl_db_filtered[
        ensembl_db_filtered['uniprot'].apply(lambda uniprot: uniprot in no_hla_uniprots['uniprot'].tolist())]

    # Add only if the ensembl didn't exist result gene list
    additional_genes = additional_genes[
        additional_genes['ensembl'].apply(lambda ensembl: not ensembl in cpdb_genes['ensembl'].tolist())
    ]

    cpdb_genes = cpdb_genes.append(additional_genes, ignore_index=True, sort=False)

    # Check if exist any duplicated ensembl
    dulicated_ensembl_genes = cpdb_genes[cpdb_genes['ensembl'].duplicated(keep=False)]

    # Remove duplicated ensembl genes if hgnc_symbol != gene_name
    cpdb_genes.drop(dulicated_ensembl_genes[
                        dulicated_ensembl_genes.apply(lambda gene: gene['hgnc_symbol'] != gene['gene_name'],
                                                      axis=1)].index, inplace=True)

    cpdb_genes = cpdb_genes.append(hla_genes, ignore_index=True, sort=False).drop_duplicates(result_columns)

    # If user_gene added, append
    cpdb_genes = cpdb_genes.append(user_gene, ignore_index=True, sort=False).drop_duplicates(result_columns,
                                                                                             keep='last')

    # Check if exist any duplicated ensembl
    # print('Duplicated ensembl genes')
    # print(len(cpdb_genes[cpdb_genes['ensembl'].duplicated()]))
    # print(cpdb_genes[cpdb_genes['ensembl'].duplicated(keep=False)].to_csv(index=False))

    return cpdb_genes[result_columns]



from cellphonedb.src.app.flask.flask_app import create_app
from cellphonedb.src.tests.cellphone_flask_test_case import CellphoneFlaskTestCase


class TestWebApiQueryInteractionCalls(CellphoneFlaskTestCase):
    def create_app(self):
        return create_app(raise_non_defined_vars=False, verbose=False)

    def test_web_api_query_interaction_call(self):
        response = self.client.get(
            '/api/v1/query/interaction/gene')

        self.assert200(response, 'Web api get interactions gene call error')
        self.assertTrue('ensembl' in str(response.data), 'ensembl column should be present in response')
        self.assertTrue('gene_name' in str(response.data), 'gene_name column shoud be present in response')
        self.assertTrue('hgnc_symbol' in str(response.data), 'hgnc_symbol should be present in response')

    def test_web_api_query_interaction_call_parameters(self):
        response = self.client.get(
            '/api/v1/query/interaction/gene?columns=ensembl,gene_name')

        self.assert200(response, 'Web api get interactions gene call with parameters error')
        self.assertTrue('ensembl' in str(response.data), 'ensembl column should be present in response')
        self.assertTrue('gene_name' in str(response.data), 'gene_name column shoud be present in response')
        self.assertTrue('hgnc_symbol' not in str(response.data), 'hgnc_symbol shouldnt be present in response')



from cellphonedb.src.core.CellphonedbSqlalchemy import CellphonedbSqlalchemy


class CellphonedbFlask(object):
    def __init__(self, app=None, config=None):
        self.app = app
        if app is not None:
            self.init_app(config)

    def init_app(self, config: dict):
        self.cellphonedb = CellphonedbSqlalchemy(config)



#!/usr/bin/env python3

import os
import click

from cellphonedb.src.api_endpoints.terminal_api.database_terminal_api_endpoints import database_terminal_commands
from cellphonedb.src.api_endpoints.terminal_api.method_terminal_api_endpoints import method_terminal_commands
from cellphonedb.src.api_endpoints.terminal_api.plot_terminal_api_endpoints import plot_terminal_commands
from cellphonedb.src.api_endpoints.terminal_api.query_terminal_api_endpoints import query_terminal_commands
from cellphonedb.src.api_endpoints.terminal_api.tools_terminal_api_endpoints import tools_terminal_commands


@click.group()
def cli():
    pass


@cli.group()
def method():
    pass


@cli.group()
def query():
    pass


@cli.group()
def database():
    pass


@cli.group()
def plot():
    pass


method.add_command(method_terminal_commands.statistical_analysis)
method.add_command(method_terminal_commands.analysis)
query.add_command(query_terminal_commands.find_interactions_by_element)
query.add_command(query_terminal_commands.get_interaction_gene)

database.add_command(database_terminal_commands.download)
database.add_command(database_terminal_commands.list_remote)
database.add_command(database_terminal_commands.list_local)
database.add_command(database_terminal_commands.generate)

plot.add_command(plot_terminal_commands.dot_plot)
plot.add_command(plot_terminal_commands.heatmap_plot)

advanced_flag_var = 'ADVANCED'

if os.getenv(advanced_flag_var, None):
    @cli.group()
    def tools():
        pass


    query.add_command(query_terminal_commands.autocomplete)
    database.add_command(database_terminal_commands.collect)
    database.add_command(database_terminal_commands.collect_generated)
    tools.add_command(tools_terminal_commands.generate_genes)
    tools.add_command(tools_terminal_commands.generate_proteins)
    tools.add_command(tools_terminal_commands.generate_complex)
    tools.add_command(tools_terminal_commands.generate_interactions)
    tools.add_command(tools_terminal_commands.filter_all)

if __name__ == '__main__':
    cli()



from cellphonedb.src.app import app_config
from cellphonedb.src.core.CellphonedbSqlalchemy import CellphonedbSqlalchemy


def create_app(verbose: bool = None, database_file=None, collecting=False):
    config = app_config.AppConfig(verbose=verbose)
    cellphone_config = config.get_cellphone_core_config()
    return CellphonedbSqlalchemy(cellphone_config, database_file, collecting)



import sys
import traceback
from typing import Optional, Any, Callable

import click
from click import Context, Argument

from cellphonedb.src.api_endpoints.terminal_api.util.choose_database import choose_database
from cellphonedb.src.app import cpdb_app
from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.core.exceptions.AllCountsFilteredException import AllCountsFilteredException
from cellphonedb.src.core.exceptions.EmptyResultException import EmptyResultException
from cellphonedb.src.core.exceptions.ThresholdValueException import ThresholdValueException
from cellphonedb.src.core.utils.subsampler import Subsampler
from cellphonedb.src.exceptions.ParseCountsException import ParseCountsException
from cellphonedb.src.exceptions.ParseMetaException import ParseMetaException
from cellphonedb.src.exceptions.ReadFileException import ReadFileException
from cellphonedb.src.local_launchers.local_method_launcher import LocalMethodLauncher


def check_subsampling_params(ctx: Context, argument: Argument, value) -> Any:
    subsampling = ctx.params.get('subsampling')

    if not subsampling and value is not None:
        tpl = 'This parameter ({}) only applies to subsampling, to enable it add `--subsampling` to your command'
        app_logger.error(tpl.format(argument.name))
        ctx.abort()

    if argument.name == 'subsampling_log' and subsampling and value is None:
        app_logger.error('''In order to perform subsampling you need to specify whether to log1p input counts or not:
            to do this specify in your command as --subsampling-log [true|false]''')
        ctx.abort()

    defaults = {
        'subsampling_num_pc': 100,
        'subsampling_num_cells': None
    }

    if subsampling and value is None:
        return defaults.get(argument.name, None)

    return value


def subsampling_options(f: Callable) -> Callable:
    options = [
        click.option('--subsampling', is_flag=True, help='Enable subsampling', is_eager=True),
        click.option('--subsampling-log', default=None, type=bool, callback=check_subsampling_params,
                     help='Enable subsampling log1p for non transformed data inputs !mandatory!'),
        click.option('--subsampling-num-pc', default=None, type=int, callback=check_subsampling_params,
                     help='Subsampling NumPC argument [100]'),
        click.option('--subsampling-num-cells', default=None, type=int, callback=check_subsampling_params,
                     help='Number of cells to subsample (defaults to a 1/3 of cells)')
    ]

    for option in reversed(options):
        f = option(f)

    return f


def common_options(f: Callable) -> Callable:
    options = [
        click.argument('meta-filename'),
        click.argument('counts-filename'),
        click.option('--counts-data', type=click.Choice(['ensembl', 'gene_name', 'hgnc_symbol']), default='ensembl'),
        click.option('--project-name', default='', type=str,
                     help='Name of the project. It creates a subfolder in output folder'),
        click.option('--threshold', default=0.1, type=float, help='% of cells expressing a gene'),
        click.option('--result-precision', default='3', type=int, help='Number of decimal digits in results [3]'),
        click.option('--output-path', default='', type=str,
                     help='Directory where the results will be allocated (the directory must exist) [out]'),
        click.option('--output-format', type=click.Choice(['txt', 'csv', 'tsv', 'tab'])),
        click.option('--means-result-name', default='means', type=str, help='Means result namefile [means]'),
        click.option('--significant-means-result-name', default='significant_means', type=str,
                     help='Significant result namefile [significant_means]'),
        click.option('--deconvoluted-result-name', default='deconvoluted',
                     help='Deconvoluted result namefile [deconvoluted]'),
        click.option('--verbose/--quiet', default=True, help='Print or hide cellphonedb logs [verbose]'),
        click.option('--database', default='latest', callback=choose_database),
        subsampling_options
    ]

    for option in reversed(options):
        f = option(f)

    return f


@click.command()
@common_options
@click.option('--debug-seed', default='-1', type=int, help='Debug random seed 0 for disable it. >=0 to set it [-1]')
@click.option('--pvalue', default=0.05, type=float, help='Pvalue threshold [0.05]')
@click.option('--pvalues-result-name', default='pvalues', type=str, help='Pvalues result namefile [pvalues]')
@click.option('--iterations', default=1000, type=int, help='Number of pvalues analysis iterations [1000]')
@click.option('--threads', default=4, type=int, help='Max of threads to process the data [4]')
def statistical_analysis(meta_filename: str,
                         counts_filename: str,
                         counts_data: str,
                         project_name: str,
                         threshold: float,
                         result_precision: int,
                         output_path: str,
                         output_format: str,
                         means_result_name: str,
                         significant_means_result_name: str,
                         deconvoluted_result_name: str,
                         verbose: bool,
                         database: Optional[str],
                         subsampling: bool,
                         subsampling_log: bool,
                         subsampling_num_pc: int,
                         subsampling_num_cells: Optional[int],
                         debug_seed: int,
                         pvalue: float,
                         pvalues_result_name: str,
                         iterations: int,
                         threads: int
                         ) -> None:
    try:

        subsampler = Subsampler(subsampling_log,
                                subsampling_num_pc,
                                subsampling_num_cells,
                                verbose) if subsampling else None

        LocalMethodLauncher(cpdb_app.create_app(verbose, database)). \
            cpdb_statistical_analysis_local_method_launcher(meta_filename,
                                                            counts_filename,
                                                            counts_data,
                                                            project_name,
                                                            iterations,
                                                            threshold,
                                                            output_path,
                                                            output_format,
                                                            means_result_name,
                                                            pvalues_result_name,
                                                            significant_means_result_name,
                                                            deconvoluted_result_name,
                                                            debug_seed,
                                                            threads,
                                                            result_precision,
                                                            pvalue,
                                                            subsampler,
                                                            )
    except (ReadFileException, ParseMetaException, ParseCountsException, ThresholdValueException,
            AllCountsFilteredException) as e:
        app_logger.error(str(e) +
                         (':' if (hasattr(e, 'description') and e.description) or (
                                 hasattr(e, 'hint') and e.hint) else '') +
                         (' {}.'.format(e.description) if hasattr(e, 'description') and e.description else '') +
                         (' {}.'.format(e.hint) if hasattr(e, 'hint') and e.hint else '')
                         )

    except EmptyResultException as e:
        app_logger.warning(str(e) +
                           (':' if (hasattr(e, 'description') and e.description) or (
                                   hasattr(e, 'hint') and e.hint) else '') +
                           (' {}.'.format(e.description) if hasattr(e, 'description') and e.description else '') +
                           (' {}.'.format(e.hint) if hasattr(e, 'hint') and e.hint else '')
                           )
    except:
        app_logger.error('Unexpected error')
        if verbose:
            traceback.print_exc(file=sys.stdout)


@click.command()
@common_options
def analysis(meta_filename: str,
             counts_filename: str,
             counts_data: str,
             project_name: str,
             threshold: float,
             result_precision: int,
             output_path: str,
             output_format: str,
             means_result_name: str,
             significant_means_result_name: str,
             deconvoluted_result_name: str,
             verbose: bool,
             database: Optional[str],
             subsampling: bool,
             subsampling_log: bool,
             subsampling_num_pc: int,
             subsampling_num_cells: Optional[int]
             ):
    try:

        subsampler = Subsampler(subsampling_log,
                                subsampling_num_pc,
                                subsampling_num_cells,
                                verbose) if subsampling else None

        LocalMethodLauncher(cpdb_app.create_app(verbose,
                                                database)).cpdb_analysis_local_method_launcher(meta_filename,
                                                                                               counts_filename,
                                                                                               counts_data,
                                                                                               project_name,
                                                                                               threshold,
                                                                                               output_path,
                                                                                               output_format,
                                                                                               means_result_name,
                                                                                               significant_means_result_name,
                                                                                               deconvoluted_result_name,
                                                                                               result_precision,
                                                                                               subsampler,
                                                                                               )
    except (ReadFileException, ParseMetaException, ParseCountsException, ThresholdValueException,
            AllCountsFilteredException) as e:
        app_logger.error(str(e) +
                         (':' if (hasattr(e, 'description') and e.description) or (
                                 hasattr(e, 'hint') and e.hint) else '') +
                         (' {}.'.format(e.description) if hasattr(e, 'description') and e.description else '') +
                         (' {}.'.format(e.hint) if hasattr(e, 'hint') and e.hint else '')
                         )

    except EmptyResultException as e:
        app_logger.warning(str(e) +
                           (':' if (hasattr(e, 'description') and e.description) or (
                                   hasattr(e, 'hint') and e.hint) else '') +
                           (' {}.'.format(e.description) if hasattr(e, 'description') and e.description else '') +
                           (' {}.'.format(e.hint) if hasattr(e, 'hint') and e.hint else '')
                           )
    except:
        app_logger.error('Unexpected error')

        if verbose:
            traceback.print_exc(file=sys.stdout)



import pandas as pd

from cellphonedb.src.core.utils import filters


def call(genes: pd.DataFrame, gene_columns: list) -> pd.DataFrame:
    genes.rename(index=str, columns={'uniprot': 'name'},
                 inplace=True)

    filters.remove_not_defined_columns(genes, gene_columns + ['name'])

    return genes



import os

from flask import Flask

this_file_dir = os.path.dirname(os.path.realpath(__file__))


data_dir = '{}/data'.format(this_file_dir)
output_dir = '{}/out'.format(this_file_dir)
downloads_dir = '{}/downloads'.format(data_dir)


def create_app():
    app = Flask(__name__)

    return app



import os
from typing import Optional

import pandas as pd

from cellphonedb.src.app.cellphonedb_app import output_test_dir, data_test_dir, cellphonedb_app
from cellphonedb.src.app.flask.flask_app import create_app
from cellphonedb.src.core.utils.subsampler import Subsampler
from cellphonedb.src.local_launchers.local_method_launcher import LocalMethodLauncher
from cellphonedb.src.tests.cellphone_flask_test_case import CellphoneFlaskTestCase
from cellphonedb.utils import dataframe_functions


class TestTerminalMethodStatisticalAnalysis(CellphoneFlaskTestCase):
    def create_app(self):
        return create_app(raise_non_defined_vars=False, verbose=False)

    def test_statistical_method__data_test__it_10__seed_0__threshold__01__precision_1(self):
        iterations = 10
        data = 'test'
        debug_seed = 0
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 1
        self._method_call(data, iterations, project_name, threshold, debug_seed, result_precision)

    def test_statistical_method__data_test__it_10__seed_0__threshold__01__precision_1_hgnc_symbol(self):
        iterations = 10
        data = 'test_custom_counts_data'
        debug_seed = 0
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 1
        self._method_call(data, iterations, project_name, threshold, debug_seed, result_precision,
                          counts_data='hgnc_symbol')

    def test_statistical_method__data_test__it_10__seed_0__threshold__01__precision_1_gene_name(self):
        iterations = 10
        data = 'test_custom_counts_data'
        debug_seed = 0
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 1
        self._method_call(data, iterations, project_name, threshold, debug_seed, result_precision,
                          counts_data='gene_name')

    def test_statistical_method__data_test__it_10__seed_0__threshold__01__precision_3(self):
        iterations = 10
        data = 'test'
        debug_seed = 0
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 3
        self._method_call(data, iterations, project_name, threshold, debug_seed, result_precision)

    def test_statistical_method_subsampled_data_test__threshold__01__precision_3__num_pc_4__num_cells_4(self):
        iterations = 10
        data = 'test_subsampled'
        debug_seed = 0
        project_name = 'test_data'
        threshold = 0.1
        result_precision = 3
        subsampler = Subsampler(False, 4, 4, debug_seed=0)
        self._method_call(data, iterations, project_name, threshold, debug_seed, result_precision, subsampler)

    def _method_call(self,
                     data: str,
                     iterations: int,
                     project_name: str,
                     threshold: float,
                     debug_seed: int,
                     result_precision: int,
                     subsampler: Optional[Subsampler] = None,
                     counts_data: str = 'ensembl'
                     ):

        result_names_as_fixture = False
        if result_names_as_fixture:
            result_deconvoluted_filename, result_means_filename, result_pvalues_filename, result_significant_means_filename = self._original_names(
                data, debug_seed, iterations, result_precision, threshold)

        else:
            result_means_filename = self._get_result_filename('means', data, iterations, debug_seed, threshold,
                                                              result_precision)
            result_pvalues_filename = self._get_result_filename('pvalues', data, iterations, debug_seed, threshold,
                                                                result_precision)
            result_significant_means_filename = self._get_result_filename('significant_means', data, iterations,
                                                                          debug_seed,
                                                                          threshold, result_precision)
            result_deconvoluted_filename = self._get_result_filename('deconvoluted', data, iterations, debug_seed,
                                                                     threshold, result_precision)

        meta_filename = os.path.realpath('{}/hi_{}_meta.txt'.format(data_test_dir, data))

        if counts_data == 'ensembl':
            counts_file = '{}/hi_{}_counts.txt'.format(data_test_dir, data)
        else:
            counts_file = '{}/hi_{}_counts_{}.txt'.format(data_test_dir, data, counts_data)

        counts_filename = os.path.realpath(counts_file)

        LocalMethodLauncher(cellphonedb_app.cellphonedb). \
            cpdb_statistical_analysis_local_method_launcher(meta_filename,
                                                            counts_filename,
                                                            counts_data,
                                                            project_name,
                                                            iterations,
                                                            threshold,
                                                            output_test_dir,
                                                            'txt',
                                                            result_means_filename,
                                                            result_pvalues_filename,
                                                            result_significant_means_filename,
                                                            result_deconvoluted_filename,
                                                            debug_seed,
                                                            result_precision=result_precision,
                                                            subsampler=subsampler,
                                                            )
        self._assert_result('means', data, iterations, project_name, result_means_filename, debug_seed, threshold,
                            result_precision)
        self._assert_result('pvalues', data, iterations, project_name, result_pvalues_filename, debug_seed, threshold,
                            result_precision)
        self._assert_result('significant_means', data, iterations, project_name, result_significant_means_filename,
                            debug_seed, threshold, result_precision)
        self._assert_result('deconvoluted', data, iterations, project_name, result_deconvoluted_filename, debug_seed,
                            threshold, result_precision)

    def _original_names(self, data, debug_seed, iterations, result_precision, threshold):
        str_threshold = ''.join(str(threshold).split('.'))
        result_means_filename = 'statistical_analysis__{}_result__' \
                                'data-{}_it-{}_seed-{}_threshold-{}_precision-{}.txt'.format('means',
                                                                                             data,
                                                                                             iterations,
                                                                                             debug_seed,
                                                                                             str_threshold,
                                                                                             result_precision)
        result_pvalues_filename = 'statistical_analysis__{}_result__' \
                                  'data-{}_it-{}_seed-{}_threshold-{}_precision-{}.txt'.format('pvalues',
                                                                                               data,
                                                                                               iterations,
                                                                                               debug_seed,
                                                                                               str_threshold,
                                                                                               result_precision)
        result_significant_means_filename = 'statistical_analysis__{}_result__' \
                                            'data-{}_it-{}_seed-{}_threshold-{}_precision-{}.txt'.format(
            'significant_means',
            data,
            iterations,
            debug_seed,
            str_threshold,
            result_precision)
        result_deconvoluted_filename = 'statistical_analysis__{}_result__' \
                                       'data-{}_it-{}_seed-{}_threshold-{}_precision-{}.txt'.format('deconvoluted',
                                                                                                    data,
                                                                                                    iterations,
                                                                                                    debug_seed,
                                                                                                    str_threshold,
                                                                                                    result_precision)
        return result_deconvoluted_filename, result_means_filename, result_pvalues_filename, result_significant_means_filename

    def _assert_result(self, filename: str,
                       data: str,
                       iterations: int,
                       project_name: str,
                       result_means_filename: str,
                       debug_seed: int,
                       threshold: float,
                       result_precision: int
                       ) -> None:
        str_threshold = ''.join(str(threshold).split('.'))

        means_test_filename = \
            'statistical_analysis__{}_result__' \
            'data-{}_it-{}_seed-{}_threshold-{}_precision-{}.txt'.format(filename,
                                                                         data,
                                                                         iterations,
                                                                         debug_seed,
                                                                         str_threshold,
                                                                         result_precision)
        original_means = pd.read_table(os.path.realpath('{}/{}'.format(data_test_dir, means_test_filename)))
        result_means = pd.read_table('{}/{}/{}'.format(output_test_dir, project_name, result_means_filename))
        self.assertTrue(dataframe_functions.dataframes_has_same_data(result_means, original_means),
                        msg='failed comparing {} with {}'.format(means_test_filename, result_means_filename))
        self.remove_file('{}/{}/{}'.format(output_test_dir, project_name, result_means_filename))

    def _get_result_filename(self,
                             base_name,
                             data: str,
                             iterations: int,
                             debug_seed: int,
                             threshold: float,
                             precision: int
                             ) -> str:
        str_threshold = ''.join(str(threshold).split('.'))

        base_filename = '{}__data-{}_it-{}_seed-{}_threshold-{}_precision-{}'.format(base_name,
                                                                                     data,
                                                                                     iterations,
                                                                                     debug_seed,
                                                                                     str_threshold,
                                                                                     precision)
        random_filename = self.get_test_filename(base_filename, 'txt')

        return random_filename



import os
import urllib.parse
from typing import Optional

import click
import pandas as pd

from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.cellphonedb_app import output_dir, data_dir
from cellphonedb.src.core.generators.complex_generator import complex_generator
from cellphonedb.src.core.generators.gene_generator import gene_generator
from cellphonedb.src.core.generators.protein_generator import protein_generator
from cellphonedb.tools import tools_helper
from cellphonedb.tools.generate_data.filters.non_complex_interactions import only_noncomplex_interactions
from cellphonedb.tools.generate_data.filters.remove_interactions import remove_interactions_in_file
from cellphonedb.tools.generate_data.getters import get_iuphar, get_imex
from cellphonedb.tools.generate_data.mergers.add_curated import add_curated
from cellphonedb.tools.generate_data.mergers.merge_interactions import merge_iuphar_imex_interactions
from cellphonedb.tools.generate_data.parsers import parse_iuphar_guidetopharmacology
from cellphonedb.tools.generate_data.parsers.parse_interactions_imex import parse_interactions_imex
from cellphonedb.utils import utils
from cellphonedb.utils.utils import _get_separator, write_to_file


@click.command()
@click.option('--user-gene', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--fetch-uniprot', is_flag=True)
@click.option('--fetch-ensembl', is_flag=True)
@click.option('--result-path', type=str, default=None)
@click.option('--project-name', type=str, default=None)
def generate_genes(user_gene: Optional[str],
                   fetch_uniprot: bool,
                   fetch_ensembl: bool,
                   result_path: str,
                   project_name: str,
                   ) -> None:
    output_path = _set_paths(result_path, project_name)

    # TODO: Add logger
    if fetch_ensembl:
        print('fetching remote ensembl data ... ', end='')
        source_url = 'http://www.ensembl.org/biomart/martservice?query={}'
        query = '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE Query><Query virtualSchemaName = "default" ' \
                'formatter = "CSV" header = "1" uniqueRows = "1" count = "" datasetConfigVersion = "0.6" >' \
                '<Dataset name = "hsapiens_gene_ensembl" interface = "default" >' \
                '<Attribute name = "ensembl_gene_id" />' \
                '<Attribute name = "ensembl_transcript_id" />' \
                '<Attribute name = "external_gene_name" />' \
                '<Attribute name = "hgnc_symbol" />' \
                '<Attribute name = "uniprotswissprot" />' \
                '</Dataset>' \
                '</Query>'

        url = source_url.format(urllib.parse.quote(query))
        ensembl_db = pd.read_csv(url)
        print('done')
    else:
        ensembl_db = utils.read_data_table_from_file(os.path.join(data_dir, 'sources/ensembl.txt'))
        print('read local ensembl file')

    # additional data comes from given file or uniprot remote url
    if fetch_uniprot:
        try:
            print('fetching remote uniprot file ... ', end='')
            source_url = 'https://www.uniprot.org/uniprot/?query=*&format=tab&force=true' \
                         '&columns=id,entry%20name,reviewed,protein%20names,genes,organism,length' \
                         '&fil=organism:%22Homo%20sapiens%20(Human)%20[9606]%22%20AND%20reviewed:yes' \
                         '&compress=yes'

            uniprot_db = pd.read_csv(source_url, sep='\t', compression='gzip')
            print('done')

        except Exception as e:
            print('Error fetching remote UniProt data, fetching local data')
            uniprot_db = pd.read_csv(os.path.join(data_dir, 'sources/uniprot.tab'), sep='\t')
            print('read local uniprot file')
    else:
        uniprot_db = utils.read_data_table_from_file(os.path.join(data_dir, 'sources/uniprot.tab'))
        print('read local uniprot file')

    ensembl_columns = {
        'Gene name': 'gene_name',
        'Gene stable ID': 'ensembl',
        'HGNC symbol': 'hgnc_symbol',
        'UniProtKB/Swiss-Prot ID': 'uniprot'
    }

    uniprot_columns = {
        'Entry': 'uniprot',
        'Gene names': 'gene_names'
    }

    result_columns = [
        'gene_name',
        'uniprot',
        'hgnc_symbol',
        'ensembl'
    ]

    ensembl_db = ensembl_db[list(ensembl_columns.keys())].rename(columns=ensembl_columns)
    uniprot_db = uniprot_db[list(uniprot_columns.keys())].rename(columns=uniprot_columns)
    hla_genes = utils.read_data_table_from_file(os.path.join(data_dir, 'sources/hla_curated.csv'))
    if user_gene:
        separator = _get_separator(os.path.splitext(user_gene)[-1])
        user_gene = pd.read_csv(user_gene, sep=separator)

    cpdb_genes = gene_generator(ensembl_db, uniprot_db, hla_genes, user_gene, result_columns)

    cpdb_genes[result_columns].to_csv('{}/{}'.format(output_path, 'gene_generated.csv'), index=False)


@click.command()
@click.argument('proteins', default='protein.csv', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.argument('genes', default='gene.csv', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.argument('complex', default='complex.csv', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--user-interactions', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--user-interactions-only', is_flag=True)
@click.option('--result-path', type=str, default=None)
@click.option('--fetch-imex', is_flag=True)
@click.option('--fetch-iuphar', is_flag=True)
@click.option('--project-name', type=str, default=None)
def generate_interactions(proteins: str,
                          genes: str,
                          complex: str,
                          user_interactions: Optional[str],
                          user_interactions_only: bool,
                          result_path: str,
                          fetch_imex: bool,
                          fetch_iuphar: bool,
                          project_name: str,
                          ) -> None:
    if user_interactions_only and not user_interactions:
        raise Exception('You need to set --user-interactions parameter')

    output_path = utils.set_paths(result_path, project_name)
    downloads_path = utils.set_paths(utils.set_paths(result_path, project_name), 'downloads')

    proteins = utils.read_data_table_from_file(proteins)
    genes = utils.read_data_table_from_file(genes)
    complexes = utils.read_data_table_from_file(complex)

    if not user_interactions_only:
        raw_imex = get_imex.call(genes, downloads_path, fetch_imex)

        interactions_to_remove = utils.read_data_table_from_file(
            os.path.join(data_dir, 'sources/excluded_interaction.csv'))
        interaction_curated = utils.read_data_table_from_file(os.path.join(data_dir, 'sources/interaction_curated.csv'))

    if user_interactions:
        separator = _get_separator(os.path.splitext(user_interactions)[-1])
        user_interactions = pd.read_csv(user_interactions, sep=separator)
        user_interactions['partner_a'] = user_interactions['partner_a'].apply(lambda x: str(x).strip())
        user_interactions['partner_b'] = user_interactions['partner_b'].apply(lambda x: str(x).strip())
        user_interactions['annotation_strategy'] = 'user_curated'

        if not 'protein_name_a' in user_interactions.columns:
            user_interactions['protein_name_a'] = ''

        if not 'protein_name_b' in user_interactions.columns:
            user_interactions['protein_name_b'] = ''

    result_columns = [
        'partner_a',
        'partner_b',
        'protein_name_a',
        'protein_name_b',
        'annotation_strategy',
        'source'
    ]
    if not user_interactions_only:
        print('Parsing IMEX file')
        imex_interactions = parse_interactions_imex(raw_imex, proteins, genes)

        print('Getting iuphar data')
        raw_iuphar = get_iuphar.call(downloads_path, fetch_iuphar)

        print('Generating iuphar interactions')
        iuphar_interactions = parse_iuphar_guidetopharmacology.call(raw_iuphar, genes, proteins)

        print('Merging iuphar/imex')
        merged_interactions = merge_iuphar_imex_interactions(iuphar_interactions, imex_interactions)

        print('Removing complex interactions')
        no_complex_interactions = only_noncomplex_interactions(merged_interactions, complexes)

        print('Removing selected interactions')
        clean_interactions = remove_interactions_in_file(no_complex_interactions, interactions_to_remove)

        print('Adding curated interaction')
        interactions_with_curated = add_curated(clean_interactions, interaction_curated)

        result = tools_helper.normalize_interactions(
            interactions_with_curated.append(user_interactions, ignore_index=True, sort=False), 'partner_a',
            'partner_b').drop_duplicates(['partner_a', 'partner_b'], keep='last')

    else:
        result = tools_helper.normalize_interactions(user_interactions, 'partner_a', 'partner_b') \
            .drop_duplicates(['partner_a', 'partner_b'], keep='last')

    result[result_columns].sort_values(['partner_a', 'partner_b']).to_csv(
        '{}/interaction_input.csv'.format(output_path), index=False)


@click.command()
@click.option('--user-protein', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--fetch-uniprot', is_flag=True)
@click.option('--result-path', type=str, default=None)
@click.option('--log-file', type=str, default='log.txt')
@click.option('--project-name', type=str, default=None)
def generate_proteins(user_protein: Optional[str],
                      fetch_uniprot: bool,
                      result_path: str,
                      log_file: str,
                      project_name: str):
    uniprot_columns = {
        'Entry': 'uniprot',
        'Entry name': 'protein_name',
    }

    # additional data comes from given file or uniprot remote url
    if fetch_uniprot:
        try:
            source_url = 'https://www.uniprot.org/uniprot/?query=*&format=tab&force=true' \
                         '&columns=id,entry%20name,reviewed,protein%20names,genes,organism,length' \
                         '&fil=organism:%22Homo%20sapiens%20(Human)%20[9606]%22%20AND%20reviewed:yes' \
                         '&compress=yes'

            uniprot_db = pd.read_csv(source_url, sep='\t', compression='gzip')

            print('read remote uniprot file')
        except Exception as e:
            print('Error fetching remote UniProt data, fetching local data')
            uniprot_db = pd.read_csv(os.path.join(data_dir, 'sources/uniprot.tab'), sep='\t')
            print('read local uniprot file')
    else:
        uniprot_db = pd.read_csv(os.path.join(data_dir, 'sources/uniprot.tab'), sep='\t')
        print('read local uniprot file')

    default_values = {
        'transmembrane': False,
        'peripheral': False,
        'secreted': False,
        'secreted_desc': pd.np.nan,
        'secreted_highlight': False,
        'receptor': False,
        'receptor_desc': pd.np.nan,
        'integrin': False,
        'other': False,
        'other_desc': pd.np.nan,
        'tags': 'To_add',
        'tags_reason': pd.np.nan,
        'tags_description': pd.np.nan,
        'pfam': pd.np.nan,
    }

    default_types = {
        'uniprot': str,
        'protein_name': str,
        'transmembrane': bool,
        'peripheral': bool,
        'secreted': bool,
        'secreted_desc': str,
        'secreted_highlight': bool,
        'receptor': bool,
        'receptor_desc': str,
        'integrin': bool,
        'other': bool,
        'other_desc': str,
        'tags': str,
        'tags_reason': str,
        'tags_description': str,
        'pfam': str,
    }

    result_columns = list(default_types.keys())

    output_path = _set_paths(result_path, project_name)
    log_path = '{}/{}'.format(output_path, log_file)
    uniprot_db = uniprot_db[list(uniprot_columns.keys())].rename(columns=uniprot_columns)
    curated_proteins = pd.read_csv(os.path.join(data_dir, 'sources/protein_curated.csv'))
    if user_protein:
        separator = _get_separator(os.path.splitext(user_protein)[-1])
        user_protein = pd.read_csv(user_protein, sep=separator)

    result = protein_generator(uniprot_db, curated_proteins, user_protein, default_values, default_types,
                               result_columns, log_path)

    result[result_columns].to_csv('{}/{}'.format(output_path, 'protein_generated.csv'), index=False)


@click.command()
@click.option('--user-complex', type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.option('--result-path', type=str, default=None)
@click.option('--log-file', type=str, default='log.txt')
@click.option('--project-name', type=str, default=None)
def generate_complex(user_complex: Optional[str],
                     result_path: str,
                     log_file: str,
                     project_name: str):
    output_path = _set_paths(result_path, project_name)
    log_path = '{}/{}'.format(output_path, log_file)

    curated_complex = pd.read_csv(os.path.join(data_dir, 'sources/complex_curated.csv'))
    if user_complex:
        separator = _get_separator(os.path.splitext(user_complex)[-1])
        user_complex = pd.read_csv(user_complex, sep=separator)

    result = complex_generator(curated_complex, user_complex, log_path)

    result.to_csv('{}/{}'.format(output_path, 'complex_generated.csv'), index=False)


@click.command()
@click.option('--input-path', type=str, default=data_dir)
@click.option('--result-path', type=str, default='filtered')
@click.option('--project-name', type=str, default=None)
def filter_all(input_path,
               result_path,
               project_name: str):
    interactions = pd.read_csv(os.path.join(input_path, 'interaction_input.csv'))
    complexes = pd.read_csv(os.path.join(input_path, 'complex_generated.csv'))
    proteins = pd.read_csv(os.path.join(input_path, 'protein_generated.csv'))
    genes = pd.read_csv(os.path.join(input_path, 'gene_generated.csv'))
    output_path = _set_paths(result_path, project_name)

    interacting_partners = pd.concat([interactions['partner_a'], interactions['partner_b']]).drop_duplicates()

    filtered_complexes = _filter_complexes(complexes, interacting_partners)
    write_to_file(filtered_complexes.sort_values('complex_name'), 'complex_input.csv', output_path=output_path)

    filtered_proteins, interacting_proteins = _filter_proteins(proteins, filtered_complexes, interacting_partners)
    write_to_file(filtered_proteins.sort_values('uniprot'), 'protein_input.csv', output_path=output_path)

    filtered_genes = _filter_genes(genes, filtered_proteins['uniprot'])
    write_to_file(filtered_genes.sort_values('gene_name'), 'gene_input.csv', output_path=output_path)

    rejected_members = interacting_partners[~(interacting_partners.isin(filtered_complexes['complex_name']) |
                                              interacting_partners.isin(filtered_proteins['uniprot']))]

    if len(rejected_members):
        app_logger.warning('There are some proteins or complexes not interacting properly: `{}`'.format(
            ', '.join(rejected_members)))


def _filter_genes(genes: pd.DataFrame, interacting_proteins: pd.Series) -> pd.DataFrame:
    filtered_genes = genes[genes['uniprot'].isin(interacting_proteins)]

    return filtered_genes


def _filter_proteins(proteins: pd.DataFrame,
                     filtered_complexes: pd.DataFrame,
                     interacting_partners: pd.DataFrame) -> (pd.DataFrame, pd.DataFrame):
    interacting_proteins = pd.concat(
        [filtered_complexes['uniprot_{}'.format(i)] for i in range(1, 5)]).drop_duplicates()

    filtered_proteins = proteins[
        proteins['uniprot'].isin(interacting_partners) | proteins['uniprot'].isin(interacting_proteins)]

    return filtered_proteins, interacting_proteins


def _filter_complexes(complexes: pd.DataFrame, interacting_partners: pd.DataFrame) -> pd.DataFrame:
    filtered_complexes = complexes[complexes['complex_name'].isin(interacting_partners)]

    return filtered_complexes


def _set_paths(output_path, subfolder):
    if not output_path:
        output_path = output_dir

    if subfolder:
        output_path = os.path.realpath(os.path.expanduser('{}/{}'.format(output_path, subfolder)))

    os.makedirs(output_path, exist_ok=True)

    if _path_is_not_empty(output_path):
        app_logger.warning(
            'Output directory ({}) exist and is not empty. Result can overwrite old results'.format(output_path))

    return output_path


def _path_is_not_empty(path):
    return bool([f for f in os.listdir(path) if not f.startswith('.')])



import io
import json
import os
import zipfile
from distutils.dir_util import copy_tree
from distutils.file_util import copy_file
from distutils.version import LooseVersion
from typing import Union

import requests

from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.cellphonedb_app import core_dir
from cellphonedb.src.app.cpdb_app import create_app
from cellphonedb.src.exceptions.NoReleasesException import NoReleasesException
from cellphonedb.src.local_launchers.local_collector_launcher import LocalCollectorLauncher

cpdb_releases = '~/.cpdb/releases'
database_file = 'cellphone.db'


def _major(version: LooseVersion) -> int:
    for component in version.version:
        if isinstance(component, int):
            return component


def _get_core_version() -> LooseVersion:
    with open(os.path.join(core_dir, 'metadata.json')) as metadata_file:
        metadata = json.load(metadata_file)
        core_version = metadata.get('database_version', 'core')

        return LooseVersion(core_version)


def _ensure_core_version_in_user_dbs():
    core_version = _get_core_version()

    user_databases_prefix = os.path.expanduser(cpdb_releases)
    dest_folder = os.path.join(user_databases_prefix, str(core_version))
    database_file_location = os.path.join(dest_folder, database_file)

    if os.path.isfile(database_file_location):
        return

    os.makedirs(dest_folder, exist_ok=True)

    copy_file(os.path.join(core_dir, database_file), dest_folder)
    copy_tree(os.path.join(core_dir, 'data'), dest_folder)


def find_database_for(value: str) -> str:
    file_candidate = os.path.expanduser(value)

    if os.path.exists(file_candidate):
        # todo: warning is perhaps not appropriate, logger doesn't allow info at this point
        app_logger.warning('User selected database `{}` is available, using it'.format(file_candidate))
        return file_candidate

    _ensure_core_version_in_user_dbs()
    user_databases_prefix = os.path.expanduser(cpdb_releases)

    if not os.path.isdir(user_databases_prefix):
        app_logger.error('No downloaded databases found, run the `database download` command from the cli first')
        exit(1)

    if value == 'latest' or not value:
        available = list_local_versions()
        latest_available = available[0]
        app_logger.warning('Latest local available version is `{}`, using it'.format(latest_available))
        value = latest_available

    downloaded_candidate = os.path.join(user_databases_prefix, value, database_file)
    valid_database = os.path.exists(downloaded_candidate)

    if valid_database:
        # todo: warning is perhaps not appropriate, logger doesn't allow info at this point
        app_logger.warning('User selected downloaded database `{}` is available, using it'.format(value))
    else:
        app_logger.warning('User selected database `{}` not available, trying to download it'.format(value))
        download_database(value)
        return find_database_for(value)

    return downloaded_candidate


def collect_database(database, output_path, **kwargs):
    database_file_path = os.path.join(output_path, database)

    app = create_app(verbose=True, database_file=database_file_path, collecting=True)
    app.database_manager.database.drop_everything()
    app.database_manager.database.create_all()

    LocalCollectorLauncher(database_file_path).all(**kwargs)


def download_database(version):
    try:
        if not version or version == 'latest':
            latest_release = _latest_release()

            version = latest_release['tag']
            zip_to_download = latest_release['url']
        else:
            releases = _list_releases()

            if version not in releases:
                app_logger.error('Unavailable version selected')
                app_logger.error('Available versions are: {}'.format(', '.join(releases.keys())))
                exit(1)

            zip_to_download = releases[version]['url']

        print('Downloading `{}` release of CellPhoneDB database'.format(version))

        output_folder = os.path.expanduser('{}/{}'.format(cpdb_releases, version))
        os.makedirs(output_folder, exist_ok=True)

        zip_response = requests.get(zip_to_download)
        with zipfile.ZipFile(io.BytesIO(zip_response.content)) as thezip:
            root_folder = thezip.namelist()[0]
            for name in thezip.namelist():
                if name.endswith('/'):
                    continue

                file_folder = os.path.dirname(name)
                file_name = os.path.basename(name)

                dest_folder = os.path.realpath(os.path.join(output_folder, os.path.relpath(file_folder, root_folder)))
                dest_file = os.path.join(dest_folder, file_name)
                os.makedirs(dest_folder, exist_ok=True)

                with thezip.open(name) as zf:
                    with open(dest_file, 'wb') as fw:
                        fw.write(zf.read())

    except NoReleasesException:
        print('There are no versions available (or connection could not be made to server to retrieve them)')
        exit(1)


def list_local_versions() -> list:
    try:
        releases_folder = os.path.expanduser(cpdb_releases)
        core = _get_core_version()

        local_versions = os.listdir(releases_folder)

        compatible_versions = [version for version in local_versions if _matching_major(core, version)]

        return sorted(compatible_versions, key=LooseVersion, reverse=True)
    except FileNotFoundError:
        return []


def list_remote_database_versions():
    try:
        releases = _list_releases()

        for idx, (_, version) in enumerate(releases.items()):
            note = ' *latest' if idx == 0 else ''
            print('version {}{}: released: {}, url: {}, compatible: {}'.format(version['tag'], note, version['date'],
                                                                               version['link'], version['compatible']))

    except NoReleasesException:
        print('There are no versions available (or connection could not be made to server to retrieve them)')
        exit(1)


def list_local_database_versions():
    releases = list_local_versions()

    if not releases:
        print('There are no versions available')
        exit(1)

    for idx, version in enumerate(releases):
        note = ' *latest' if idx == 0 else ''
        print('version {}{}'.format(version, note))


def _list_releases() -> dict:
    try:
        releases = _github_query('releases')

        if not releases:
            raise NoReleasesException()

        return _format_releases(*releases)

    except requests.exceptions.ConnectionError:
        raise NoReleasesException()


def _latest_release():
    try:
        compatible_versions = {key: value for key, value in _list_releases().items() if value['compatible']}

        if not compatible_versions:
            raise NoReleasesException()

        latest = sorted(compatible_versions, key=LooseVersion, reverse=True)[0]

        return compatible_versions[latest]

    except requests.exceptions.ConnectionError:
        raise NoReleasesException()


def _github_query(kind) -> Union[dict, list]:
    queries = {
        'releases': 'https://api.github.com/repos/{}/{}/releases'.format('Teichlab', 'cellphonedb-data'),
    }

    query = queries[kind]

    if not query:
        raise Exception('unexpected query to github')

    response = requests.get(query, headers={'Accept': 'application/vnd.github.v3+json'})

    if response.status_code != 200:
        raise NoReleasesException()

    return response.json()


def _format_releases(*releases) -> dict:
    core_version = _get_core_version()

    return {item['tag_name']: _format_release(item, core_version) for item in releases}


def _format_release(item: dict, core: LooseVersion) -> dict:
    tag_name = item['tag_name']

    return {'url': item['zipball_url'],
            'tag': tag_name,
            'date': item['published_at'],
            'link': item['html_url'],
            'compatible': _matching_major(core, tag_name)
            }


def _matching_major(core, candidate):
    return _major(LooseVersion(candidate)) == _major(core)



from cellphonedb.src.app.flask.flask_app import create_app
from cellphonedb.src.app.cellphonedb_app import output_test_dir
from cellphonedb.src.local_launchers.local_exporter_launcher import LocalExporterLauncher

import os

from cellphonedb.src.tests.cellphone_flask_test_case import CellphoneFlaskTestCase


class TestExporterCalls(CellphoneFlaskTestCase):

    def test_all_exporter_generators(self):
        exporters = ['complex', 'interaction', 'protein', 'gene', 'protein_complex_cellphonedb']

        for exporter in exporters:
            self.assert_file_exist(exporter)

    def assert_file_exist(self, method_name, message='', expected_extension='csv', expected_namefile=''):
        if not expected_namefile:
            expected_namefile = '{}'.format(method_name)

        namefile = self.get_test_filename(expected_namefile, expected_extension)
        if not message:
            message = 'File {} didnt exist'.format(namefile)

        getattr(LocalExporterLauncher(), method_name)(namefile, output_test_dir)

        path_file = '{}/{}'.format(output_test_dir, namefile)
        self.assertTrue(os.path.isfile(path_file), message)
        self.remove_file(path_file)

    def setUp(self):
        self.reset_db()
        self.populate_db()

    def create_app(self):
        return create_app(environment='test', raise_non_defined_vars=False)



import pandas as pd


def get_multidatas_from_interactions(interactions: pd.DataFrame, multidatas: pd.DataFrame) -> pd.DataFrame:
    interaction_multidatas = pd.DataFrame(
        interactions['multidata_1_id'].append(interactions['multidata_2_id']).drop_duplicates(),
        columns=['id_multidata'])

    result = multidatas.merge(interaction_multidatas, on='id_multidata')

    return result



from unittest import TestCase

import pandas as pd

from cellphonedb.utils import dataframe_functions


class TestDataframeSameData(TestCase):
    def test_compare_empty(self):
        self.assertTrue(dataframe_functions.dataframes_has_same_data(pd.DataFrame(), pd.DataFrame()))

    def test_equal(self):
        dataframe1 = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})
        dataframe2 = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})

        self.assertTrue(dataframe_functions.dataframes_has_same_data(dataframe1, dataframe2))

    def test_different(self):
        dataframe1 = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})
        dataframe2 = pd.DataFrame({'col1': [5, 2], 'col2': [3, 4]})

        self.assertFalse(dataframe_functions.dataframes_has_same_data(dataframe1, dataframe2))

    def test_equal_unsorted(self):
        dataframe1 = pd.DataFrame({'col1': [2, 1], 'col2': [4, 3]})
        dataframe2 = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})

        self.assertTrue(dataframe_functions.dataframes_has_same_data(dataframe1, dataframe2))

    def test_different_unsorted(self):
        dataframe1 = pd.DataFrame({'col1': [2, 1], 'col2': [4, 5]})
        dataframe2 = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})

        self.assertFalse(dataframe_functions.dataframes_has_same_data(dataframe1, dataframe2))

    def test_equal_unsorted_columns(self):
        dataframe1 = pd.DataFrame({'col2': [3, 4], 'col1': [1, 2]})
        dataframe2 = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})

        self.assertTrue(dataframe_functions.dataframes_has_same_data(dataframe1, dataframe2))

    def test_different_unsorted_columns(self):
        dataframe1 = pd.DataFrame({'col2': [3, 4], 'col1': [1, 2]})
        dataframe2 = pd.DataFrame({'col1': [1, 2], 'col2': [3, 5]})

        self.assertFalse(dataframe_functions.dataframes_has_same_data(dataframe1, dataframe2))



from flask import Flask
from flask_restful import Api

from cellphonedb.src.api_endpoints.web_api import routes
from cellphonedb.src.app import app_config
from cellphonedb.src.app.cellphonedb_app import cellphonedb_app


def create_app(environment=None, support=None, load_defaults=None, raise_non_defined_vars=True, verbose=None):
    app = Flask(__name__)
    config = app_config.AppConfig(environment, support, load_defaults, raise_non_defined_vars, verbose=verbose)

    cellphone_config = config.get_cellphone_core_config()

    cellphonedb_app.init_app(cellphone_config)

    flask_config = config.flask_config()
    app.config.from_mapping(flask_config)
    app.url_map.strict_slashes = False

    api = Api(app, prefix=flask_config['API_PREFIX'])

    routes.add(api, '/v1')

    return app



import pandas as pd

from cellphonedb.utils import dataframe_format


# TODO: Move to helper
def call(complexes: pd.DataFrame, multidatas: pd.DataFrame, complex_compositions: pd.DataFrame,
         proteins: pd.DataFrame) -> pd.DataFrame:
    complex_complete = pd.merge(complexes, multidatas, left_on='complex_multidata_id',
                                right_on='id_multidata')

    composition = []
    for complex_index, complex in complexes.iterrows():
        complex_complex_composition = complex_compositions[
            complex_compositions['complex_multidata_id'] == complex['complex_multidata_id']]

        protein_index = 1
        complex_proteins = {
            'complex_multidata_id': complex['complex_multidata_id'],
            'uniprot_1': None, 'protein_1_gene_name': None, 'tags_1': None, 'protein_name_1': None,
            'tags_description_1': None, 'tags_reason_1': None,
            'uniprot_2': None, 'protein_2_gene_name': None, 'tags_2': None, 'protein_name_2': None,
            'tags_description_2': None, 'tags_reason_2': None,
            'uniprot_3': None, 'protein_3_gene_name': None, 'tags_3': None, 'protein_name_3': None,
            'tags_description_3': None, 'tags_reason_3': None,
            'uniprot_4': None, 'protein_4_gene_name': None, 'tags_4': None, 'protein_name_4': None,
            'tags_description_4': None, 'tags_reason_4': None,
        }
        for index, complex_composition in complex_complex_composition.iterrows():
            proteine_name = \
                multidatas[multidatas['id_multidata'] == complex_composition['protein_multidata_id']][
                    'name'].values[0]
            complex_proteins['uniprot_%i' % protein_index] = proteine_name

            selected_protein = proteins[proteins['name'] == proteine_name]
            protein_name = selected_protein['protein_name'].values[0]
            complex_proteins['protein_%i_gene_name' % protein_index] = protein_name
            complex_proteins['protein_name_%i' % protein_index] = protein_name
            complex_proteins['tags_%i' % protein_index] = selected_protein['tags'].values[0]
            complex_proteins['tags_description_%i' % protein_index] = selected_protein['tags_description'].values[0]
            complex_proteins['tags_reason_%i' % protein_index] = selected_protein['tags_reason'].values[0]
            protein_index += 1


        composition.append(complex_proteins)

    composition_df = pd.DataFrame(composition)
    complex_complete = pd.merge(complex_complete, composition_df, on='complex_multidata_id')

    complex_complete.rename({'name': 'complex_name'}, axis=1, inplace=1)

    return complex_complete[
        ['complex_name', 'uniprot_1', 'uniprot_2', 'uniprot_3', 'uniprot_4', 'transmembrane', 'peripheral', 'secreted',
         'secreted_desc', 'secreted_highlight', 'receptor', 'receptor_desc', 'integrin', 'other', 'other_desc',
         'pdb_id', 'pdb_structure', 'stoichiometry', 'comments_complex']]



import pandas as pd
import numpy as np

from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.database import DatabaseManager
from cellphonedb.src.core.exceptions.ThresholdValueException import ThresholdValueException
from cellphonedb.src.core.methods import cpdb_analysis_method, cpdb_statistical_analysis_method
from cellphonedb.src.core.preprocessors import method_preprocessors
from cellphonedb.src.core.utils.subsampler import Subsampler
from cellphonedb.src.exceptions.ParseCountsException import ParseCountsException


class MethodLauncher:
    def __init__(self, database_manager: DatabaseManager, default_threads: int, separator: str = '|'):
        self.database_manager = database_manager
        self.default_threads = default_threads
        self.separator = separator

    def __getattribute__(self, name):
        method = object.__getattribute__(self, name)
        if hasattr(method, '__call__'):
            core_logger.info('Launching Method {}'.format(name))

        return method

    def get_multidatas_from_string(self, string: str) -> pd.DataFrame:
        multidatas = self.database_manager.get_repository('multidata').get_multidatas_from_string(string)
        return multidatas

    def cpdb_statistical_analysis_launcher(self,
                                           raw_meta: pd.DataFrame,
                                           counts: pd.DataFrame,
                                           counts_data: str,
                                           iterations: int,
                                           threshold: float,
                                           threads: int,
                                           debug_seed: int,
                                           result_precision: int,
                                           pvalue: float,
                                           subsampler: Subsampler = None,
                                           ) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame):

        if threads < 1:
            core_logger.info('Using Default thread number: %s' % self.default_threads)
            threads = self.default_threads

        if threshold < 0 or threshold > 1:
            raise ThresholdValueException(threshold)

        meta = method_preprocessors.meta_preprocessor(raw_meta)
        counts = self._counts_validations(counts, meta)

        if subsampler is not None:
            counts = subsampler.subsample(counts)
            meta = meta.filter(items=(list(counts)), axis=0)

        interactions = self.database_manager.get_repository('interaction').get_all_expanded(include_gene=False)
        genes = self.database_manager.get_repository('gene').get_all_expanded()
        complex_composition = self.database_manager.get_repository('complex').get_all_compositions()
        complex_expanded = self.database_manager.get_repository('complex').get_all_expanded()

        deconvoluted, means, pvalues, significant_means = \
            cpdb_statistical_analysis_method.call(meta,
                                                  counts,
                                                  counts_data,
                                                  interactions,
                                                  genes,
                                                  complex_expanded,
                                                  complex_composition,
                                                  iterations,
                                                  threshold,
                                                  threads,
                                                  debug_seed,
                                                  result_precision,
                                                  pvalue,
                                                  self.separator)

        return pvalues, means, significant_means, deconvoluted

    def cpdb_method_analysis_launcher(self,
                                      raw_meta: pd.DataFrame,
                                      counts: pd.DataFrame,
                                      counts_data: str,
                                      threshold: float,
                                      result_precision: int,
                                      subsampler: Subsampler = None,
                                      ) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame):

        if threshold < 0 or threshold > 1:
            raise ThresholdValueException(threshold)
        meta = method_preprocessors.meta_preprocessor(raw_meta)

        counts = self._counts_validations(counts, meta)

        if subsampler is not None:
            counts = subsampler.subsample(counts)
            meta = meta.filter(items=list(counts), axis=0)

        interactions = self.database_manager.get_repository('interaction').get_all_expanded(include_gene=False)
        genes = self.database_manager.get_repository('gene').get_all_expanded()
        complex_composition = self.database_manager.get_repository('complex').get_all_compositions()
        complex_expanded = self.database_manager.get_repository('complex').get_all_expanded()

        means, significant_means, deconvoluted = cpdb_analysis_method.call(
            meta,
            counts,
            counts_data,
            interactions,
            genes,
            complex_expanded,
            complex_composition,
            self.separator,
            threshold,
            result_precision)

        return means, significant_means, deconvoluted

    @staticmethod
    def _counts_validations(counts: pd.DataFrame, meta: pd.DataFrame) -> pd.DataFrame:
        if not len(counts.columns):
            raise ParseCountsException('Counts values are not decimal values', 'Incorrect file format')
        try:
            if np.any(counts.dtypes.values != np.dtype('float32')):
                counts = counts.astype(np.float32)  # type: pd.DataFrame
        except:
            raise ParseCountsException
        meta.index = meta.index.astype(str)
        if np.any(~meta.index.isin(counts.columns)):
            raise ParseCountsException('Some cells in meta didnt exist in counts',
                                       'Maybe incorrect file format')
        if np.any(~counts.columns.isin(meta.index)):
            counts = counts.loc[:, counts.columns.isin(meta.index)].copy()
        return counts



import pandas as pd

from cellphonedb.src.core.core_logger import core_logger
from cellphonedb.src.core.database import DatabaseManager


class Repository():
    def __init__(self, database_manager: DatabaseManager):
        self.database_manager = database_manager

    # TODO: Needs refactor
    @staticmethod
    def _blend_column(original_df: pd.DataFrame, multidata_df: pd.DataFrame, original_column_name: list,
                      db_column_name: list,
                      table_name: str, number: int) -> pd.DataFrame:
        """

        :type original_df: pd.DataFrame
        :type multidata_df: pd.DataFrame
        :type original_column_name: str
        :type db_column_name: str
        :type table_name: str
        :type number: int
        :rtype: pd.DataFrame
        """
        interaction_df = pd.merge(original_df, multidata_df, left_on=original_column_name, right_on=db_column_name,
                                  indicator=True, how='outer')
        interaction_df.rename(index=str, columns={'id_%s' % table_name: '%s_%s_id' % (table_name, number)},
                              inplace=True)

        interaction_df = interaction_df[
            (interaction_df['_merge'] == 'both') | (interaction_df['_merge'] == 'left_only')]
        interaction_df.rename(index=str,
                              columns={'_merge': '_merge_%s' % number, db_column_name: db_column_name + '_%s' % number},
                              inplace=True)

        return interaction_df

    # TODO: Needs refactor
    @staticmethod
    def blend_dataframes(left_df, left_column_names, right_df, db_column_name, db_table_name, quiet=False):
        result_df = left_df.copy()

        if not quiet and db_column_name in left_df.columns:
            core_logger.debug('WARNING | BLENDING: column "%s" already exists in orginal df' % (db_column_name))

        unique_slug = '_EDITNAME'
        unique_original_column_names = [("%s%s" % (column_name, unique_slug)) for column_name in left_column_names]

        result_df.rename(index=str, columns=dict(zip(left_column_names, unique_original_column_names)),
                         inplace=True)

        not_existent_proteins = []

        for i in range(0, len(unique_original_column_names)):
            result_df = Repository._blend_column(result_df, right_df, unique_original_column_names[i],
                                                 db_column_name,
                                                 db_table_name, i + 1)

            not_existent_proteins = not_existent_proteins + \
                                    result_df[result_df['_merge_%s' % (i + 1)] == 'left_only'][
                                        unique_original_column_names[i]].drop_duplicates().tolist()
        not_existent_proteins = list(set(not_existent_proteins))

        for i in range(1, len(unique_original_column_names) + 1):
            result_df = result_df[(result_df['_merge_%s' % i] == 'both')]

        result_df.drop(['_merge_%s' % merge_column for merge_column in
                        range(1, len(unique_original_column_names) + 1)] + unique_original_column_names, axis=1,
                       inplace=True)

        if len(left_column_names) == 1:
            result_df.rename(index=str, columns={'%s_1' % db_column_name: db_column_name,
                                                 '%s_1_id' % db_table_name: '%s_id' % db_table_name}, inplace=True)

        if not quiet and not_existent_proteins:
            core_logger.warning('WARNING | BLENDING: THIS %s DIDNT EXIST IN %s' % (db_column_name, db_table_name))
            core_logger.warning(not_existent_proteins)

        return result_df



import pandas as pd

from cellphonedb.src.core.exceptions.ProcessMetaException import ProcessMetaException


def meta_preprocessor(meta_raw: pd.DataFrame) -> pd.DataFrame:
    meta_raw.columns = map(str.lower, meta_raw.columns)
    try:
        if 'cell' in meta_raw and 'cell_type' in meta_raw:
            meta = meta_raw[['cell', 'cell_type']]
            meta.set_index('cell', inplace=True, drop=True)
            return meta

        if type(meta_raw.index) == pd.core.indexes.multi.MultiIndex:
            raise ProcessMetaException

        elif 'cell_type' in meta_raw:
            meta = meta_raw[['cell_type']]
            if type(meta_raw.index) == pd.core.indexes.range.RangeIndex:
                meta.set_index(meta_raw.iloc[:, 0], inplace=True)
                meta.index.name = 'cell'
                return meta

            if type(meta_raw.index) == pd.core.indexes.base.Index:
                meta.index.name = 'cell'
                return meta

        meta = pd.DataFrame(data={'cell_type': meta_raw.iloc[:, 1]})
        meta.set_index(meta_raw.iloc[:, 0], inplace=True)
        meta.index.name = 'cell'
        meta.index = meta.index.astype(str)
        return meta

    except:
        raise ProcessMetaException



import pandas as pd


def bring_columns_to_start(columns, dataframe):
    """

    :type columns: list
    :type dataframe: pd.DataFrame
    :rtype: pd.DataFrame
    """
    column_headers = list(dataframe.columns.values)
    for column in reversed(columns):
        column_headers.insert(0, column_headers.pop(column_headers.index(column)))

    result = dataframe[column_headers].copy()
    return result


def bring_columns_to_end(columns, dataframe):
    """

    :type columns: list
    :type dataframe: pd.DataFrame
    :rtype: pd.DataFrame
    """

    column_headers = list(dataframe.columns.values)
    for column in columns:
        column_headers.append(column_headers.pop(column_headers.index(column)))

    result = dataframe[column_headers].copy()
    return result


def change_column_suffix(dataframe: pd.DataFrame, old_suffix: str, new_suffix: str) -> pd.DataFrame:
    """
    Renames multiple columns suffixes
    """
    column_names = list(dataframe.columns.values)

    renamed_columns = {}
    for column_name in column_names:
        if column_name.endswith(old_suffix):
            new_name = column_name[:-len(old_suffix)] + new_suffix
            renamed_columns[column_name] = new_name
    interaction_preprocess_collector.py
    dataframe = dataframe.rename(index=str, columns=renamed_columns)
    return dataframe



from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.cellphonedb_app import data_dir, cellphonedb_app
from cellphonedb.src.app.cpdb_app import create_app
from cellphonedb.utils import utils


class LocalCollectorLauncher(object):

    def __init__(self, database_file=None):
        self.database_file = database_file

    def __getattr__(self, method_name):
        def wrapper(namefile='', data_path=''):
            app_logger.info('Collecting {}'.format(method_name))
            if not namefile:
                namefile = '{}_input.csv'.format(method_name)

            if not data_path:
                data_path = data_dir

            data = utils.read_data_table_from_file('{}/{}'.format(data_path, namefile))

            if self.database_file:
                getattr(create_app(True, self.database_file, True).collect, method_name)(data)
            else:
                getattr(cellphonedb_app.cellphonedb.collect, method_name)(data)

        return wrapper

    def all(self, protein_filename='', gene_filename='', complex_filename='', interaction_filename='', data_path=''):
        self.protein(protein_filename, data_path)
        self.gene(gene_filename, data_path)
        self.complex(complex_filename, data_path)
        self.interaction(interaction_filename, data_path)



import pandas as pd
from flask_testing import TestCase

from cellphonedb.src.app.cellphonedb_app import cellphonedb_app
from cellphonedb.src.app.app_logger import app_logger
from cellphonedb.src.app.flask.flask_app import create_app

complex_entries = [
    {
        'data': {
            "name": "aVb3 complex",
            "transmembrane": True,
            "peripheral": False,
            "secreted": False,
            "secreted_desc": None,
            "secreted_highlight": False,
            "receptor": False,
            "receptor_desc": None,
            "integrin": True,
            "other": False,
            "other_desc": None,
            "pdb_id": "1jv2",
            "pdb_structure": "TRUE",
            "stoichiometry": "ITGAV;ITGB3",
            "comments_complex": "Well known integrin combination"
        },
        'composition': ["P06756", "P05106"]
    },
    {
        'data': {
            "name": "a2Bb3 complex",
            "receptor": False,
            "receptor_desc": None,
            "other": False,
            "other_desc": None,
            "secreted_highlight": False,
            "secreted_desc": None,
            "transmembrane": True,
            "secreted": False,
            "peripheral": False,
            "pdb_structure": "TRUE",
            "pdb_id": "1kup",
            "stoichiometry": "ITGA2B;ITGB3",
            "comments_complex": "Well known integrin combination",
            "integrin": True

        },
        'composition':
            ["P08514", "P05106"]
    },

    {
        'data': {
            "name": "IL2 receptor_HA",
            "transmembrane": True,
            "peripheral": False,
            "secreted": False,
            "secreted_desc": None,
            "secreted_highlight": False,
            "receptor": True,
            "receptor_desc": "Cytokine receptor IL2 family",
            "integrin": False,
            "other": False,
            "other_desc": None,
            "pdb_id": "2b5i",
            "pdb_structure": "binding",
            "stoichiometry": "IL2;IL2RA;IL2RB;IL2RG",
            "comments_complex": "A high affinity dimer, an intermediate affinity monomer (beta subunit). The high and intermediate affinity forms also associate with a gamma subunit."
        },
        'composition':
            ["P01589", "P14784", "P31785"]
    }
]
protein_entries = [
    {
        "name": "P39019",
        "protein_name": "RS19_HUMAN",
        "transmembrane": False,
        "peripheral": False,
        "secreted": False,
        "secreted_desc": None,
        "secreted_highlight": False,
        "receptor": False,
        "receptor_desc": None,
        "integrin": False,
        "other": False,
        "other_desc": None,
        "tags": "To_add",
        "tags_reason": None,
        "tags_description": None,

    }, {

        "name": "P54760",
        "protein_name": "EPHB4_HUMAN",
        "transmembrane": True,
        "peripheral": False,
        "secreted": False,
        "secreted_desc": None,
        "secreted_highlight": False,
        "receptor": True,
        "receptor_desc": None,
        "integrin": False,
        "other": False,
        "other_desc": None,
        "tags": "To_comment",
        "tags_reason": "Adhesion_add",
        "tags_description": None

    }, {
        "name": "P52799",
        "protein_name": "EFNB2_HUMAN",
        "transmembrane": True,
        "peripheral": False,
        "secreted": False,
        "secreted_desc": None,
        "secreted_highlight": False,
        "receptor": True,
        "receptor_desc": None,
        "integrin": False,
        "other": False,
        "other_desc": None,
        "tags": "To_add",
        "tags_reason": None,
        "tags_description": "ligandIUPHAR",
    },
    {
        "name": "P39059",
        "protein_name": "COFA1_HUMAN",
        "transmembrane": False,
        "peripheral": False,
        "secreted": True,
        "secreted_desc": None,
        "secreted_highlight": False,
        "receptor": False,
        "receptor_desc": None,
        "integrin": False,
        "other": True,
        "other_desc": "Collagen",
        "tags": "To_add",
        "tags_reason": None,
        "tags_description": "Collagen"
    },
]
gene_entries = [
    {
        "ensembl": "ENSG00000275555",
        "gene_name": "DLL1",
        "name": "O00548"
    },
    {
        "ensembl": "ENSG00000169306",
        "gene_name": "IL1RAPL1",
        "name": "Q9NZN1"
    },
    {
        "ensembl": "ENSG00000204642",
        "gene_name": "HLA-F",
        "name": "P30511"
    }
]
interaction_entries = [
    {
        "id_cp_interaction": "CPI-CS0A66DB1CA",
        "name_1": "CD8 receptor",
        "name_2": "P06239",
        "source": "1q69, PDB partially",
        "annotation_strategy": "curated",
    },
    {
        "id_cp_interaction": "CPI-SS085EE60B1",
        "name_1": "P01137",
        "name_2": "Q03167",
        "source": "uniprot",
        "annotation_strategy": "curated",
    },
    {
        "id_cp_interaction": "CPI-SS056BE1011",
        "name_1": "O00421",
        "name_2": "Q99731",
        "source": None,
        "annotation_strategy": "guidetopharmacology.org",
    },
    {
        "id_cp_interaction": "CPI-SS03165AD8C",
        "name_1": "O00590",
        "name_2": "P13500",
        "source": "PMID: 24218476",
        "annotation_strategy": "curated",
    },
    {
        "id_cp_interaction": "CPI-SS0F972435E",
        "name_1": "P35916",
        "name_2": "P49767",
        "source": "uniprot",
        "annotation_strategy": "curated",
    },
    {
        "id_cp_interaction": "CPI-SS027E57635",
        "name_1": "Q8NFJ6",
        "name_2": "Q9HC23",
        "annotation_strategy": "guidetopharmacology.org",
        "source": None
    },
    {
        "id_cp_interaction": "CPI-SS0DBA4D668",
        "name_1": "P27487",
        "name_2": "P48061",
        "annotation_strategy": "curated",
        "source": "PMID: 24218476"
    }
]


class TestValidatorDatabaseRandomEntries(TestCase):
    def test_interaction(self):

        interaction_df = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'interaction').get_all_expanded()

        data_not_match = False

        for interaction in interaction_entries:
            db_interaction = interaction_df
            non_match_properties = []
            for column_name in interaction:
                if interaction[column_name] == None:
                    db_interaction = db_interaction[pd.isnull(db_interaction[column_name])]
                else:
                    db_interaction = db_interaction[db_interaction[column_name] == interaction[column_name]]

                if len(db_interaction) < 1:
                    non_match_properties.append(column_name)
            if (len(db_interaction) < 1):
                app_logger.warning('Failed cheking Interaction:')
                app_logger.warning('Expected data:')
                app_logger.warning(interaction)
                app_logger.warning('Non Match properties')
                app_logger.warning(non_match_properties)
                data_not_match = True

        self.assertFalse(data_not_match, 'Some Interactions doesnt match')

    def test_gene(self):

        dataframe = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'gene').get_all_expanded()

        data_not_match = False

        for gene in gene_entries:
            db_gene = dataframe

            for column_name in gene:
                if gene[column_name] == None:
                    db_gene = db_gene[pd.isnull(db_gene[column_name])]
                else:
                    db_gene = db_gene[db_gene[column_name] == gene[column_name]]

            if (len(db_gene) < 1):
                app_logger.warning('Failed cheking Gene:')
                app_logger.warning('Expected data:')
                app_logger.warning(gene)
                data_not_match = True

        self.assertFalse(data_not_match, 'Some Gene doesnt match')

    def test_protein(self):

        dataframe = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'protein').get_all_expanded()

        data_not_match = False

        for protein in protein_entries:
            db_protein = dataframe[dataframe['name'] == protein['name']]

            if db_protein.empty:
                print('Protein {} dindt exist'.format(protein['name']))
                data_not_match = True
                continue

            for column_name in protein:
                if db_protein[column_name].iloc[0] != protein[column_name]:
                    app_logger.warning('Failed checking column \'%s\' of multidata/protein with name \'%s\'' % (
                        column_name, protein['name']))
                    app_logger.warning('Expected value: %s' % protein[column_name])
                    app_logger.warning('Database value: %s' % db_protein[column_name].iloc[0])
                    app_logger.warning('---')
                    data_not_match = True

        self.assertFalse(data_not_match, 'Some proteins doesnt match or doesnt exist')

    def test_complex_composition_table(self):
        df_multidata = cellphonedb_app.cellphonedb.database_manager.get_repository('multidata').get_all()
        df_complex_composition = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'complex').get_all_compositions()

        number_compositions_not_match = False
        some_protein_didnt_exists = False
        some_protein_not_part_of_complex = False
        some_complex_not_exist = False

        for complex in complex_entries:
            try:
                db_complex_id = df_multidata[df_multidata['name'] == complex['data']['name']]['id_multidata'].iloc[0]
            except IndexError:
                print('Complex {} didnt exist'.format(complex['data']['name']))
                some_complex_not_exist = True
                continue

            if len(df_complex_composition[df_complex_composition['complex_multidata_id'] != db_complex_id]) == len(
                    complex['composition']):
                app_logger.warning('Failed checking number of complex_composition with name \'%s\'' % (
                    complex['data']['name']))
                app_logger.warning('Expected value: %s' % len(
                    df_complex_composition[df_complex_composition['complex_multidata_id'] == db_complex_id]))
                app_logger.warning('Database value: %s' % len(complex['composition']))
                app_logger.warning('---')
                number_compositions_not_match = True

            for protein_name in complex['composition']:
                db_complex_composition_ids = \
                    df_complex_composition[df_complex_composition['complex_multidata_id'] == db_complex_id][
                        'protein_multidata_id'].tolist()

                composition_multidata_id = df_multidata[df_multidata['name'] == protein_name]['id_multidata']

                if not len(composition_multidata_id):
                    app_logger.warning('Failed finding protein \'%s\' in multidata from complex name \'%s\'' % (
                        protein_name, complex['data']['name']))
                    some_protein_didnt_exists = True
                    continue

                if composition_multidata_id.iloc[0] not in db_complex_composition_ids:
                    app_logger.warning('Failed finding protein \'%s\' in composition from complex name \'%s\'' % (
                        protein_name, complex['data']['name']))
                    some_protein_not_part_of_complex = True

        self.assertFalse(number_compositions_not_match, 'Number of complex composition doesnt match')
        self.assertFalse(some_protein_didnt_exists, 'Some complex_composition proteins doesnt match')
        self.assertFalse(some_protein_not_part_of_complex, 'Complex_composition proteins doesnt match')
        self.assertFalse(some_complex_not_exist, 'Some Complex not exist')

    def test_complex(self):

        dataframe = cellphonedb_app.cellphonedb.database_manager.get_repository(
            'complex').get_all_expanded()

        data_not_match = False

        for complex in complex_entries:
            db_complex = dataframe[dataframe['name'] == complex['data']['name']]

            for complex_data in complex['data']:
                if db_complex[complex_data].iloc[0] != complex['data'][complex_data]:
                    app_logger.warning('Failed checking column \'%s\' of multidata/complex with name \'%s\'' % (
                        complex_data, complex['data']['name']))
                    app_logger.warning('Expected value: %s' % complex['data'][complex_data])
                    app_logger.warning('Database value: %s' % db_complex[complex_data].iloc[0])
                    app_logger.warning('---')
                    data_not_match = True

        self.assertFalse(data_not_match, 'Some complex doesnt match')

    def create_app(self):
        return create_app(raise_non_defined_vars=False, verbose=False)

    def setUp(self):
        self.client = self.app.test_client()



import os
from functools import wraps
from typing import Optional

import pandas as pd

from cellphonedb.src.exceptions.MissingPlotterFunctionException import MissingPlotterFunctionException
from cellphonedb.src.exceptions.MissingR import MissingR
from cellphonedb.src.exceptions.RRuntimeException import RRuntimeException
from cellphonedb.utils.utils import _get_separator


def ensure_R_setup():
    from rpy2 import situation
    try:
        if not situation.get_r_home() or not situation.r_version_from_subprocess():
            raise MissingR()

    except MissingR as e:
        raise e


def _ensure_path_exists(path: str) -> None:
    expanded_path = os.path.expanduser(path)

    if not os.path.exists(expanded_path):
        os.makedirs(expanded_path)


def with_r_setup(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        ensure_R_setup()

        from rpy2.rinterface_lib.embedded import RRuntimeError
        from rpy2 import robjects

        return f(*args, **kwargs, robjects=robjects, r_runtime_error=RRuntimeError)

    return wrapper


@with_r_setup
def heatmaps_plot(*,
                  meta_file: str,
                  pvalues_file: str,
                  output_path: str,
                  count_name: str,
                  log_name: str,
                  count_network_filename: str,
                  interaction_count_filename: str,
                  pvalue: float,
                  robjects,
                  r_runtime_error: Exception
                  ) -> None:
    _ensure_path_exists(output_path)
    meta_file_separator = _get_separator(os.path.splitext(meta_file)[-1])
    pvalues_file_separator = _get_separator(os.path.splitext(pvalues_file)[-1])
    this_file_dir = os.path.dirname(os.path.realpath(__file__))
    robjects.r.source(os.path.join(this_file_dir, 'R/plot_heatmaps.R'))
    available_names = list(robjects.globalenv.keys())
    count_filename = os.path.join(output_path, count_name)
    log_filename = os.path.join(output_path, log_name)
    plot_function = 'heatmaps_plot'

    if plot_function in available_names:
        function_name = plot_function
    else:
        raise MissingPlotterFunctionException()

    plotter = robjects.r[function_name]

    count_network_separator = _get_separator(os.path.splitext(count_network_filename)[-1])
    interaction_count_separator = _get_separator(os.path.splitext(interaction_count_filename)[-1])

    count_network_filename = os.path.join(output_path, count_network_filename)
    interaction_count_filename = os.path.join(output_path, interaction_count_filename)

    try:
        plotter(meta_file=meta_file,
                pvalues_file=pvalues_file,
                count_filename=count_filename,
                log_filename=log_filename,
                count_network_filename=count_network_filename,
                interaction_count_filename=interaction_count_filename,
                count_network_separator=count_network_separator,
                interaction_count_separator=interaction_count_separator,
                meta_sep=meta_file_separator,
                pvalues_sep=pvalues_file_separator,
                pvalue=pvalue,
                )
    except r_runtime_error as e:
        raise RRuntimeException(e)


@with_r_setup
def dot_plot(*,
             means_path: str,
             pvalues_path: str,
             output_path: str,
             output_name: str,
             robjects,
             r_runtime_error: Exception,
             rows: Optional[str] = None,
             columns: Optional[str] = None
             ) -> None:
    _ensure_path_exists(output_path)
    pvalues_separator = _get_separator(os.path.splitext(pvalues_path)[-1])
    means_separator = _get_separator(os.path.splitext(means_path)[-1])
    output_extension = os.path.splitext(output_name)[-1].lower()
    filename = os.path.join(output_path, output_name)

    means_df = pd.read_csv(means_path, sep=means_separator)
    n_rows, n_cols = means_df.shape
    n_cols -= 11

    n_rows, selected_rows = selected_items(rows, n_rows)
    n_cols, selected_columns = selected_items(columns, n_cols)

    this_file_dir = os.path.dirname(os.path.realpath(__file__))
    robjects.r.source(os.path.join(this_file_dir, 'R/plot_dot_by_column_name.R'))
    available_names = list(robjects.globalenv.keys())
    plot_function = 'dot_plot'

    if plot_function in available_names:
        function_name = plot_function
    else:
        raise MissingPlotterFunctionException()

    plotter = robjects.r[function_name]

    try:
        plotter(selected_rows=selected_rows,
                selected_columns=selected_columns,
                filename=filename,
                width=int(5 + max(3, n_cols * 0.8)),
                height=int(5 + max(5, n_rows * 0.5)),
                means_path=means_path,
                pvalues_path=pvalues_path,
                means_separator=means_separator,
                pvalues_separator=pvalues_separator,
                output_extension=output_extension
                )
    except r_runtime_error as e:
        raise RRuntimeException(e)


@with_r_setup
def selected_items(selection: Optional[str], size, *, robjects, r_runtime_error):
    if selection is not None:
        df = pd.read_csv(selection, header=None)
        names = df[0].tolist()

        from rpy2.robjects.vectors import StrVector
        selected = StrVector(names)
        size = len(names)
    else:
        selected = robjects.NULL

    return size, selected



import pandas as pd


def call(interactions_expanded: pd.DataFrame) -> pd.DataFrame:
    interactions_expanded.rename(index=str, columns={'name_1': 'partner_a',
                                                     'name_2': 'partner_b',
                                                     'protein_name_1': 'protein_name_a',
                                                     'protein_name_2': 'protein_name_b'},
                                 inplace=True)

    interactions_expanded.sort_values('id_cp_interaction', inplace=True)
    return interactions_expanded[
        ['id_cp_interaction', 'partner_a', 'partner_b', 'protein_name_a', 'protein_name_b', 'annotation_strategy',
         'source']]



class EmptyResultException(Exception):
    def __init__(self, description: str = None, hint: str = None):
        super(EmptyResultException, self).__init__('Result is empty')
        self.description = description
        self.hint = hint



from cellphonedb.tools.generate_data.filters import remove_genes
from cellphonedb.tools.generate_data.mergers import mergers_genes
from cellphonedb.tools.validators import gene_validators
from cellphonedb.utils import utils


def remove_genes_in_file(gene_base_filename: str,
                         remove_genes_filename: str,
                         result_filename: str = 'gene_filtered.csv',
                         ) -> None:

    gene_base_data = utils.read_data_table_from_file(gene_base_filename)
    remove_genes_data = utils.read_data_table_from_file(remove_genes_filename)

    genes_filtered = remove_genes.remove_genes_in_file(gene_base_data, remove_genes_data)

    genes_filtered.to_csv(result_filename, index=False)


def generate_genes_from_uniprot_ensembl_db(uniprot_db_filename: str,
                                           ensembl_db_filename: str,
                                           proteins_filename: str,
                                           result_filename: str = 'gene_uniprot_ensembl_merged.csv',
                                           ) -> None:

    uniprots = utils.read_data_table_from_file(uniprot_db_filename)
    ensembls = utils.read_data_table_from_file(ensembl_db_filename)
    proteins = utils.read_data_table_from_file(proteins_filename)

    result = mergers_genes.merge_genes_from_uniprot_ensembl_db(ensembls, proteins, uniprots)

    result.to_csv(result_filename, index=False)


def add_hla_genes(gene_base_filename: str,
                  hla_genes_filename: str,
                  result_filename: str = 'gene_hla.csv',
                  ) -> None:

    gene_base = utils.read_data_table_from_file(gene_base_filename)
    hla_genes = utils.read_data_table_from_file(hla_genes_filename)

    genes_merged = gene_base.append(hla_genes, ignore_index=True)

    genes_merged.to_csv(result_filename, index=False)


def validate_gene_list(gene_filename: str) -> None:
    genes = utils.read_data_table_from_file(gene_filename)

    if gene_validators.validate_genes(genes):
        print('GENE LIST IS VALID')

    else:
        print('GENE LIST IS NOT VALID')



from sqlalchemy import Column, Integer, ForeignKey

from cellphonedb.src.core.database.sqlalchemy_models import Base


class ComplexComposition(Base):
    __tablename__ = 'complex_composition_table'
    id_complex_composition = Column(Integer, nullable=False, primary_key=True)

    complex_multidata_id = Column(Integer, ForeignKey('multidata_table.id_multidata'), nullable=False)
    protein_multidata_id = Column(Integer, ForeignKey('multidata_table.id_multidata'), nullable=False)
    total_protein = Column(Integer)



import pandas as pd

from cellphonedb.src.app.flask.flask_app import create_app
from cellphonedb.src.app.cellphonedb_app import cellphonedb_app
from cellphonedb.src.local_launchers.local_collector_launcher import LocalCollectorLauncher
from cellphonedb.src.tests.cellphone_flask_test_case import CellphoneFlaskTestCase


class TestCollectionCalls(CellphoneFlaskTestCase):

    def test_collect_data(self):
        cellphonedb_app.cellphonedb.database_manager.database.drop_everything()
        cellphonedb_app.cellphonedb.database_manager.database.create_all()

        self.check_proteins()
        self.check_genes()
        self.check_complex()
        self.check_interaction()

    def check_proteins(self):
        self.collect_data('protein')
        self.assert_number_data('protein')

        proteins_expected = pd.read_csv('{}/{}'.format(self.fixtures_dir(), 'collect_protein.csv'))
        multidatas_db = cellphonedb_app.cellphonedb.database_manager.get_repository('multidata').get_all()

        self.assertEqual(len(proteins_expected), len(multidatas_db),
                         'Database collected multidata (from proteins) didnt match')

    def check_genes(self):
        self.collect_data('gene')
        self.assert_number_data('gene')

    def check_complex(self):
        self.collect_data('complex')
        self.assert_number_data('complex')

    def check_interaction(self):
        self.collect_data('interaction')
        self.assert_number_data('interaction')

    def assert_number_data(self, name):
        namefile = 'collect_{}.csv'.format(name)

        db_data = cellphonedb_app.cellphonedb.database_manager.get_repository(name).get_all()

        expected_data = pd.read_csv('{}/{}'.format(self.fixtures_dir(), namefile))

        self.assertEqual(len(expected_data), len(db_data), 'Database collected {} didnt match'.format(name))

    def collect_data(self, name):
        namefile = 'collect_{}.csv'.format(name)
        getattr(LocalCollectorLauncher(), name)(namefile, self.fixtures_dir())

    def create_app(self):
        return create_app(environment='test', raise_non_defined_vars=False)



import json
import os
from datetime import datetime
from json import JSONDecodeError

import pandas as pd
import tqdm


def interaction_exist(interaction, interactions, interaction_1_key='uniprot_1', interaction_2_key='uniprot_2'):
    """
    Checks if interaction already exists in first dataframe.
    """

    if len(interactions[(interactions[interaction_1_key] == interaction[interaction_1_key]) & (
            interactions[interaction_2_key] == interaction[interaction_2_key])]):
        return True

    if len(interactions[(interactions[interaction_2_key] == interaction[interaction_1_key]) & (
            interactions[interaction_1_key] == interaction[interaction_2_key])]):
        return True

    return False


def sort_interactions_partners_alphabetically(interactions: pd.DataFrame,
                                              names: tuple = ('partner_1', 'partner_2')) -> pd.DataFrame:
    """
    Sorts interactions partners alphabetically:
    ie:
    A->B        A->B
    B->C        B->C
    D->A   =>   A->D
    B->A        A->B
    E->D        D->E
    """
    interactions = interactions.copy()

    def alphabetic_sort(interaction: pd.Series):
        return pd.Series({
            names[0]: min([interaction[names[0]], interaction[names[1]]]),
            names[1]: max([interaction[names[0]], interaction[names[1]]])
        })

    interactions[[names[0], names[1]]] = interactions.apply(alphabetic_sort, axis=1)

    return interactions


def normalize_interactions(interactions, interaction_1_key='protein_1', interaction_2_key='protein_2'):
    """
    Permutes all inversed interactions:
    ie:
    A->B        A->B
    B->A        A->B
    B->A   =>   A->B
    A->B        A->B
    B->A        A->B

    """
    return sort_interactions_partners_alphabetically(interactions, (interaction_1_key, interaction_2_key))


def add_to_meta(file: str, meta_path: str):
    if not os.path.exists(meta_path):
        meta = {}
    else:
        try:
            with open(meta_path, 'r') as metafile:
                meta = json.load(metafile)
        except JSONDecodeError:
            meta = {}

    with open(meta_path, 'w') as metafile:
        meta[file] = {'date': datetime.now().strftime('%Y%m%d')}
        json.dump(meta, metafile, indent=2)
